# AddOn Tile Proc (c) 2014 Atom

# Wheel formula.
# -(len*offset)/wheel_radius

import bpy, random, math

from mathutils import Vector, Matrix

from .util import returnNameForNumber
from .util import returnCurveName
from .util import returnIndexedName
from .util import returnItemName
from .util import arclength
from .util import returnFollowPathConstraint
from .util import curveFromEndPoints
from .util import returnActionFromObject
from .util import returnMapNeighborTypes
from .util import returnTileNeighborTypes
from .util import eraseAllKeyframes
from .util import removeCurveFromMemory
from .util import returnSelfDeterminedByNeighbors

def cloneBirdRigWithOptions(scene, source_rig_name, rig_index, x, y, source_body_name, material_name):
    # "rigticle" a particle rig.
    #print ("cloneBirdRigWithOptions")
    new_rig_name = ""
    ob_arm_source = bpy.data.objects.get(source_rig_name)
    if ob_arm_source != None:
        ob_body_source = bpy.data.objects.get(source_body_name)
        if ob_body_source != None:
            # We have all the required parts to make a clone.
            #new_rig_name = returnIndexedName(source_rig_name, rig_index)
            new_rig_name = returnItemName(source_rig_name, x, y, rig_index)
            #print ("New rig name [%s]" % new_rig_name)
            ob_arm_new = bpy.data.objects.get(new_rig_name)
            if ob_arm_new == None:
                # I don't exist yet, time to create.
                ob_arm_new = ob_arm_source.copy()
                ob_arm_new.name = new_rig_name
                #print(ob_arm_new.name)
                scene.objects.link(ob_arm_new)
            else:
                print ("cloneBirdRigWithOptions: No ob_arm_new")
                
            # Clone the bird body.    
            #new_body_name = returnIndexedName(("%s_BODY" % source_rig_name), rig_index)
            new_body_name = returnItemName(("%s_BODY" % source_rig_name),x, y, rig_index)
            ob_body_new = bpy.data.objects.get(new_body_name)
            if ob_body_new == None:
                # I don't exist yet, time to create.
                ob_body_new = ob_body_source.copy()
                ob_body_new.name = new_body_name
                ob_body_new.location = (0.,0.0,0.0)
                ob_body_new.rotation_euler = (0.0,0.0,0.0)
                ob_body_new.scale = (1.0,1.0,1.0)
            else:
                ob_body_new.data = ob_body_source.data
            try:
                scene.objects.link(ob_body_new)
            except:
                pass
            ob_body_new.parent = ob_arm_new
            # Make the Armature modifier point to this new rig object.
            for mod in ob_body_new.modifiers:
                if mod.type == 'ARMATURE':
                    # Assume, only expecting one armature on this object.
                    mod.object = ob_arm_new
            # Assign material.
            mat = bpy.data.materials.get(material_name)
            if mat != None:
                ob_body_new.material_slots[0].material = mat
        else:
            print ("cloneBirdRigWithOptions: No ob_body_source")
    else:
        print ("cloneBirdRigWithOptions: No arm_source")
    return new_rig_name  
                  
def cloneVehicleRigWithOptions(scene, source_rig_name, rig_index, source_wheels_name, source_body_name, material_name, clone_lights = False):
    # "rigticle" a particle rig.
    #print ("cloneVehicleRigWithOptions")
    new_rig_name = ""
    ob_arm_source = bpy.data.objects.get(source_rig_name)
    if ob_arm_source != None:
        ob_wheel_source = bpy.data.objects.get(source_wheels_name)
        if ob_wheel_source != None:
            ob_body_source = bpy.data.objects.get(source_body_name)
            if ob_body_source != None:
                # We have all the required parts to make a clone.
                new_rig_name = returnIndexedName(source_rig_name, rig_index)
                #print ("New rig name [%s]" % new_rig_name)
                ob_arm_new = bpy.data.objects.get(new_rig_name)
                if ob_arm_new == None:
                    # I don't exist yet, time to create.
                    ob_arm_new = ob_arm_source.copy()
                    ob_arm_new.name = new_rig_name
                    #print(ob_arm_new.name)
                    scene.objects.link(ob_arm_new)
                else:
                    print ("cloneVehicleRigWithOptions: No ob_arm_new")
                    
                # Clone children of armature if they are of the light type.
                if clone_lights:
                    for i,child in enumerate(ob_arm_source.children):
                        
                        if child.type == 'LAMP':
                            print(child)
                            new_light_name = returnIndexedName(("%s" % child.name), rig_index) 
                            ob_light_new = bpy.data.objects.get(new_light_name)
                            if ob_light_new == None:
                                # I don't exist yet, time to create.
                                ob_light_new = child.copy()
                                ob_light_new.name = new_light_name
                                ob_light_new.location = child.location
                                ob_light_new.rotation_euler = child.rotation_euler
                                ob_light_new.scale = child.scale
                            else:
                                ob_light_new.data = child.data
                            try:
                                scene.objects.link(ob_light_new)
                            except:
                                pass
                            ob_light_new.parent = ob_arm_new

                # Clone the vehicle body.    
                new_body_name = returnIndexedName(("%s_BODY" % source_rig_name), rig_index)
                ob_body_new = bpy.data.objects.get(new_body_name)
                if ob_body_new == None:
                    # I don't exist yet, time to create.
                    ob_body_new = ob_body_source.copy()
                    ob_body_new.name = new_body_name
                    ob_body_new.location = (0.,0.0,0.0)
                    ob_body_new.rotation_euler = (0.0,0.0,0.0)
                    ob_body_new.scale = (1.0,1.0,1.0)
                else:
                    ob_body_new.data = ob_body_source.data
                try:
                    scene.objects.link(ob_body_new)
                except:
                    pass
                ob_body_new.parent = ob_arm_new
                # Make the Armature modifier point to this new rig object.
                for mod in ob_body_new.modifiers:
                    if mod.type == 'ARMATURE':
                        # Assume, only expecting one armature on this object.
                        mod.object = ob_arm_new
                # Assign material.
                mat = bpy.data.materials.get(material_name)
                if mat != None:
                    ob_body_new.material_slots[0].material = mat
                
                # Clone the vehicle wheels.    
                new_wheels_name = returnIndexedName(("%s_WHEELS" % source_rig_name), rig_index)
                ob_wheels_new = bpy.data.objects.get(new_wheels_name)
                if ob_wheels_new == None:
                    # I don't exist yet, time to create.
                    ob_wheels_new = ob_wheel_source.copy()
                    ob_wheels_new.name = new_wheels_name
                    ob_wheels_new.location = (0.,0.0,0.0)
                    ob_wheels_new.rotation_euler = (0.0,0.0,0.0)
                    ob_wheels_new.scale = (1.0,1.0,1.0)
                else:
                    ob_wheels_new.data = ob_wheel_source.data
                try:
                    scene.objects.link(ob_wheels_new)
                except:
                    pass
                ob_wheels_new.parent = ob_arm_new
                # Make the Armature modifier point to this new rig object.
                for mod in ob_wheels_new.modifiers:
                    if mod.type == 'ARMATURE':
                        # Assume, only expecting one armature on this object.
                        mod.object = ob_arm_new
            else:
                print ("cloneVehicleRigWithOptions: No ob_body_source")
        else:
            print ("cloneVehicleRigWithOptions: No ob_wheel_source")
    else:
        print ("cloneVehicleRigWithOptions: No arm_source")
    return new_rig_name
        
def returnWheelRotationFromDistance(armatureName, boneName, wheelRadius, currentFrame):
    result = 0.0
    ob = bpy.data.objects.get(armatureName)
    if ob != None:
        # Scan the constraint list to determine what curve we are currently traveling along.
        for i,cn in enumerate(ob.constraints):
            if cn.influence > 0.0:
                # This constraint is active.
                if cn.target != None:
                    ob_curve = bpy.data.objects.get(cn.target.name)
                    if ob_curve != None:
                        curve_length = arclength([ob_curve])
                        wheel_bone = ob.pose.bones.get(boneName)
                        if wheel_bone != None:
                            current_rotation = wheel_bone.rotation_euler[0]    # Fetch our current value.
                            
                            # A wheel has roughly four states: stopped, rolling, speeding up and cruising.
                            # Examine the offset_factor in the past and future to see if the wheel is even moving.
                            offset_past = 0.0
                            offset_future = 0.0
                            offset_now = cn.offset_factor
                            action = returnActionFromObject(ob)
                            if action != None:
                                for j,fcurve in enumerate(action.fcurves):
                                    if fcurve.data_path.find(cn.name) != -1:
                                        # This curve belongs to the constraint that is currently influencing the rig.
                                        if (fcurve.data_path.find('offset_factor') != -1):
                                            # This is the fcurve for the offset_factor of the currently active constraint.
                                            offset_past = fcurve.evaluate(currentFrame -1)
                                            offset_future = fcurve.evaluate(currentFrame +1)
                                            
                                            #distance_traveled = (offset_future - offset_past) * curve_length
                                            distance_traveled = offset_now * curve_length
                                            if distance_traveled != 0.0:
                                                # The wheel is moving.
                                                circumference = (2 * wheelRadius) * math.pi 
                                                num_rots = distance_traveled / circumference
                                                #result = ( (-((distance_traveled / circumference) * 360)) + current_rotation)
                                                result = -((distance_traveled / circumference) * 360) - (i * 1.0) # last part helps with cross over when transitioning between curves. (1.0 should take into account wheel diameter but HARD CODED for now).
                                                #print ("[%s] traveling via constraint #%i [%s]" % (boneName,i,cn.name))
                                                #print(cn.name, currentFrame, num_rots, distance_traveled, offset_past, offset_future)
                                            else:
                                                # The wheel is stopped, return current value.
                                                result = current_rotation
                                        break                                            
                            else:
                                print ("returnWheelRotationFromDistance: no actions discovered on [%s]" % armatureName)
                        else:
                            print ("returnWheelRotationFromDistance: pose bone [%s] is not valid." % boneName)
                    else:
                        print ("returnWheelRotationFromDistance: target [%s] does not exist." % cn.target.name)
                else:
                    print ("returnWheelRotationFromDistance: constraint [%s] has no target specified." % cn.name)
            else:
                pass
                #print ("[%s] has no influence." % cn.name)
    else:
        print ("returnWheelRotationFromDistance: no action discovered on [%s]" % armatureName)
    return result

def add_curve(scene, curve_name, points):
    curve_data_name = ("cu_%s" % curve_name)
    road_path = bpy.data.curves.get(curve_data_name)
    if road_path == None:
        # Make a new one.
        road_path = bpy.data.curves.new(curve_data_name, type='CURVE')
    else:
        #Reuse this existing curve, but we must remove all of the current splines.
        for old_spline in road_path.splines:
            try:
                road_path.splines.remove(old_spline)
            except:
                print("add_curve: ERROR trying to remove old_spline")
    
    spline = road_path.splines.new('BEZIER')
    spline.bezier_points.add(len(points)-1)

    curvobj = bpy.data.objects.get(curve_name)
    if curvobj == None:
        curvobj = bpy.data.objects.new(curve_name,road_path)
    else:
        old_curve_name = curvobj.data.name
        if old_curve_name != curve_data_name:
            # The names are different, time to link to the new datablock.
            curvobj.data = road_path
            removeCurveFromMemory(old_curve_name)

    for i in range(len(points)):
        spline.bezier_points[i].co = points[i]
        spline.bezier_points[i].handle_right_type = 'AUTO'
        spline.bezier_points[i].handle_left_type = 'AUTO'
    try:
        scene.objects.link(curvobj)
    except:
        # Alreadylinked.
        pass

    #curvobj.matrix_world = matrix
    curvobj.show_x_ray = True
    return curvobj

def generateFlyingBird (context, c, x, y, layer, area_map, entry):
    scene = context.scene
    ob_parent = context.object
    w = len(area_map)		# width
    h = len(area_map[0])	# height
    unit_size = bpy.context.user_preferences.addons['tile_proc'].preferences.unit_size	# The size, in Blender units, that each grid element represents in the scene.
    width_div2 = unit_size * (w/2)
    height_div2 = unit_size * (h/2)
    rnd_seed = layer.rnd_seed
    source_body_name = 'crow'		#HARD CODED name reference to scene object.
    material_name = 'mat_crow'		#HARD CODED name reference to scene object.
    rig_name = cloneBirdRigWithOptions(scene, layer.bird_rig_name, c, x, y, source_body_name, material_name)   # Clone the rig, and body mesh.
    if len(rig_name) == 0:
        return
    rnd_seed += c
    
    ob = bpy.data.objects.get(rig_name)
    if ob != None:
        ob.hide = layer.hide_construction
        #ob.hide = True
        ob.parent = ob_parent
        # Generate a random value for the coordinate.
        co = [0,0,0]
        noise = 0.68
        for i in range(3):
            co[i] += noise*(random.random()*2-1)
            
        # Make a randomized circle path for the bird to follow.
        points = []
        curve_vector_set = []
        is_high_z = False
        if (entry.ob_name.find('high') != -1): is_high_z = True
        if (entry.ob_name.find('bridge') != -1): is_high_z = True
        if is_high_z:
            z = 9.4			# HARD CODED z-axis height for bird paths.
        else:
            z = 5.6			# HARD CODED z-axis height for bird paths.
        coordinate_x = (unit_size * x) - (unit_size * co[0]) - width_div2
        coordinate_y = (-unit_size * y) - (unit_size * co[1]) - height_div2
        radius = 1.0
        noise = 0.23
        rnd_z = noise*(random.random()*2-1)
        rnd_x = noise*(random.random()*unit_size - width_div2)
        rnd_y = noise*(random.random()*unit_size - height_div2)
        curve_vector_set.append((0,rnd_y,z - (unit_size * rnd_z)))
        rnd_z = noise*(random.random()*2-1)
        rnd_x = noise*(random.random()*unit_size - width_div2)
        rnd_y = noise*(random.random()*unit_size - height_div2)
        curve_vector_set.append((rnd_x,0,z - (unit_size * rnd_z)))
        rnd_z = noise*(random.random()*2-1)
        rnd_x = noise*(random.random()*unit_size - width_div2)
        rnd_y = noise*(random.random()*unit_size - height_div2)
        curve_vector_set.append((0,-rnd_y,z - (unit_size * rnd_z)))
        rnd_z = noise*(random.random()*2-1)
        rnd_x = noise*(random.random()*unit_size - width_div2)
        rnd_y = noise*(random.random()*unit_size - height_div2)
        curve_vector_set.append((-rnd_x,0,z - (unit_size * rnd_z)))
        for v1 in curve_vector_set:
            points.append(v1)
        if len(points) > 0:
            # Use points to construct a new curve object. In our case a circle randomly over whatever x,y tile this is.
            path_bird_name = "path_bird_%s" % rig_name
            ob_path_circle = add_curve(scene, path_bird_name, points)
            ob_path_circle.data.splines[0].use_cyclic_u = True
            ob_path_circle.data.dimensions = '3D'
            #ob_path_circle.location = ((unit_size * x) - width_div2, (unit_size * y) - height_div2, 0)
            ob_path_circle.location.x = unit_size * x - (unit_size * (w/2))
            ob_path_circle.location.y = -unit_size * y - (-unit_size * (h/2))
            #ob_path_circle.parent = ob_parent

        ob.constraints.clear()	# Remove constraints at the object level, the pose mode constraints are not affected by this.
        # Remove all pose bone animation.
        #ad = ob.animation_data
        #if ad != None:
        #   ob.animation_data_clear()

        rnd = (unit_size/2) - int(random.random() * unit_size)
        start_frame = 1 + rnd
        end_frame = (c+1) * (layer.bird_speed + rnd)
        if ob_path_circle:
            ob_path_circle.hide = layer.hide_construction
            #ob_path_circle.parent = ob
            # For the next few lines of code we are going to override the User Preferences for keyframe interpolation type.
            # Note: This technique does not seem to work in 2.72 release..?
            keyInterp = bpy.context.user_preferences.edit.keyframe_new_interpolation_type
            keyHandle = bpy.context.user_preferences.edit.keyframe_new_handle_type
            bpy.context.user_preferences.edit.keyframe_new_interpolation_type ='LINEAR'
            bpy.context.user_preferences.edit.keyframe_new_handle_type ='VECTOR'
            
            eraseAllKeyframes(ob_path_circle)
            # Assign a follow path constraint the rig.
            cn = returnFollowPathConstraint(path_bird_name, ob, ob_path_circle, 1.0)
            if cn != None:
                s = start_frame
                e = end_frame
                for c in range(42):		# Loop path many times.
                    # Insert two linear keyframes to animate along the path.
                    cn.offset_factor = 1.0
                    cn.keyframe_insert(data_path='offset_factor', frame=(int(s)))
                    cn.offset_factor = 0.0
                    cn.keyframe_insert(data_path='offset_factor', frame=(int(e)))
                    s = e + 1
                    e += end_frame

            # Restore User Preferences to what it was.
            bpy.context.user_preferences.edit.keyframe_new_interpolation_type = keyInterp
            bpy.context.user_preferences.edit.keyframe_new_handle_type = keyHandle   
            
            # Fetch the f-Curve that we just created via the frame_insert method.
            try:
                ad_source = ob
                action_name = ad_source.animation_data.action.name
            except:
                action_name = ""
            try:
                action = bpy.data.actions[action_name]
                can_proceed = True
            except:
                can_proceed = False
            if can_proceed == True:
                for i,fcurve in enumerate(action.fcurves):
                    if fcurve.data_path.find('offset_factor') != -1:
                        # Make this fcurve use LINEAR interpolation.
                        fcurve.extrapolation = 'LINEAR'
                        for point in fcurve.keyframe_points:
                            point.interpolation = 'LINEAR'
              
def generateDrivingVehicle (scene, index, layer, area_map, maze_map, solution_map, solution_coordinates, grid, collection):
    # Begin generation of road path using maze algorithms.
    print("\n")

    # Maze Generation Parameters:
    w = len(area_map)		#12                  	 # width
    h = len(area_map[0])	#12                  	 # height
    branch_rate = layer.branch_rate			#3       # Lower for straight mazes, higher for curve mazes. (-10 to 10 range).
    rnd_seed = layer.rnd_seed				#831     # Pick a new number to make a new version of the maze.
    area = layer.area						#1.0     # 1.0 makes the area map a solid. Lower values introduces random holes with 0.0 having no solid area.
    area_center_x = layer.area_center_x		#0.0     # Offset the area map to purposely introduce wide sections of empty space. (-1.0 to 1.0)     
    area_center_y = layer.area_center_y		#0.0     # Offset the area map to purposely introduce wide sections of empty space. (-1.0 to 1.0) 

    unit_size = bpy.context.user_preferences.addons['tile_proc'].preferences.unit_size	# The size, in Blender units, that each grid element represents in the scene.
    curve_subdiv = layer.road_resolution	#7        # The resolution of each generated curve. Increase if you use noise.
    noise = layer.road_noise				#0.07     # deviation from a straight line. 
    threshold = layer.road_threshold		#0.65     # threshold for road bumpiness.
        
    car_count = layer.vehicle_count			#9
    source_rig_name =  layer.vehicle_rig_name			#'rig_car'                         # An object name of armature type.                
    source_wheels_name = 'rig_car_wheel_set'  			# An object name of vertex mapped mesh type.   HARD CODED reference to scene object.            

    # These two lists must be synchronized both in length and content.
    vehicle_body_sources = ['vehicle_body_normal', 'vehicle_body_hatchback', 'vehicle_body_micro_cargo', 'vehicle_body_micro_transport', 'vehicle_body_pickup_small', 'vehicle_body_micro','vehicle_body_coupe','vehicle_body_station_wagon','vehicle_body_mpv','vehicle_body_minibus','vehicle_body_van']
    vehicle_wheel_bases = [(0.09, 0.09, 0.04, -0.04), (0.13, 0.09, 0, 0), (0.05, 0.09, 0, 0), (0.05, 0.09, 0, 0), (0.09, 0.09, 0.036, -0.036), (0.19, 0.0, 0.04, -0.04), (-.0768, 0.012, 0.03, -0.03), (-0.1, 0.09, 0.036, -0.036), (-0.1, 0.09, 0.036, -0.036), (-0.235, 0.09, 0, 0), (-0.235, 0.09, 0, 0)]
    #vehicle_wheels_sources = ['vehicle_wheels_normal', 'vehicle_wheels_hatchback', 'vehicle_wheels_micro_cargo', 'vehicle_wheels_micro_transport', 'vehicle_wheels_pickup_small']

    # The possible material options for whichever body type is chosen.
    vehicle_material_options = ['carmaterial_silver','carmaterial_blue', 'carmaterial_citrus', 'carmaterial_green', 'carmaterial_orange', 'carmaterial_red', 'carmaterial_violet']

    #for c in range(car_count):
    c = index
    l = len(vehicle_material_options)
    ndx = int(random.random() * l)
    material_name = vehicle_material_options[ndx]
    
    l = len(vehicle_body_sources)
    ndx = int(random.random() * l)
    source_body_name = vehicle_body_sources[ndx]
    wheel_base_rear, wheel_base_front, wheel_base_left, wheel_base_right = vehicle_wheel_bases[ndx]
    
    rig_name = cloneVehicleRigWithOptions(scene,source_rig_name, c, source_wheels_name, source_body_name, material_name, layer.vehicle_lights)   # Clone the rig, wheels and car body mesh.
    if len(rig_name) == 0:
        return
    rnd_seed += c
    
    # 1.) Create an area map. (0.0-1.0) 0.0 is 100% hollow and 1.0 is 100% solid.
    # 2.) Create a maze based upon input parameters. Use the area map hollow and solid grid elements to indicate where the maze can carve into.
    # 3.) Solve the maze producing a single path through the maze.
   
    #area_map = returnRndAreaMap(w, h, area, area_center_x, area_center_y, rnd_seed)
    #maze_map = returnMazeMap(w, h, branch_rate, area_map,rnd_seed)  
    #solution_map, solution_coordinates = solveMaze(maze_map)

    # Examine your system console to see this output.
    #printGrid(area_map,"Area Map.")
    #printGrid(maze_map,"Maze Map.")
    #printGrid(solution_map,"Solution Map.")

    # 4.) Convert each element in the solution map into a curve.
    x_last = -1
    y_last = -1
    curve_vector_set = []

    # An empty for all the road parts to reside under.
    road_part_parent_name = returnIndexedName(("mt_%s" % rig_name), c)
    mt_road_parts = bpy.data.objects.get(road_part_parent_name)
    if mt_road_parts == None:
        # Does not exist, create.
        mt_road_parts = bpy.data.objects.new(road_part_parent_name, None)
    else:
        # Already exists, let's unlink any children objects.
        for ob_child in mt_road_parts.children:
            if ob_child.type == 'CURVE':
                try:
                    scene.objects.unlink(ob_child)
                except:
                    pass
    try:
        scene.objects.link(mt_road_parts)
    except:
        pass
    
    bpy.data.objects[rig_name].parent = mt_road_parts

    for i,(x,y) in enumerate(solution_coordinates):
        l = len(solution_coordinates)
        if (i == (l-1)):
            # Start and end points should fetch from the maze map.
            top, bottom, left, right = returnMapNeighborTypes(maze_map, x, y)
        else:
            # Intermediate steps should fetch from the solution map.
            top, bottom, left, right = returnMapNeighborTypes(solution_map, x, y)
        connection_type = returnSelfDeterminedByNeighbors(top,bottom,left,right)

        # Needs to know the direction to travel from the current location.
        x_dir = 0
        y_dir = 0
        if top == '.':
            # Possible direction candidate.
            if (y-1) != y_last:
                # Not the same as last coordinate I guess this ok considering our solution map contains only one way through the maze.
                y_dir =-1
        if bottom == '.':
            # Possible direction candidate.
            if (y+1) != y_last:
                # Not the same as last coordinate I guess this ok considering our solution map contains only one way through the maze.
                y_dir =1
        if left == '.':
            # Possible direction candidate.
            if (x-1) != x_last:
                # Not the same as last coordinate I guess this ok considering our solution map contains only one way through the maze.
                x_dir =-1 
        if right == '.':
            # Possible direction candidate.
            if (x+1) != x_last:
                # Not the same as last coordinate I guess this ok considering our soultion map contains only one way through the maze.
                x_dir =1
                
        x_last = x
        y_last = y
        '''
        skip_crossroads = False
        if x == 0: skip_crossroads = True
        if x == (w-1): skip_crossroads = True
        if y == 0: skip_crossroads = True
        if y == (h-1): skip_crossroads = True
        if skip_crossroads == False:
            # Crossroads intersection.
            if connection_type == 'XX' and (entry.id_top in match_types) and (entry.id_bottom in match_types) and (entry.id_left in match_types) and (entry.id_right in match_types): 
                tile_id = ndx
                break
        # T intersections.	
        if connection_type == 'TT' and (entry.id_top not in match_types) and (entry.id_bottom in match_types) and (entry.id_left in match_types) and (entry.id_right in match_types): 
            tile_id = ndx
            break
        if connection_type == 'TB' and (entry.id_top in match_types) and (entry.id_bottom not in match_types) and (entry.id_left in match_types) and (entry.id_right in match_types): 
            tile_id = ndx
            break
        if connection_type == 'TL' and (entry.id_top in match_types) and (entry.id_bottom in match_types) and (entry.id_left not in match_types) and (entry.id_right in match_types): 
            tile_id = ndx
            break
        if connection_type == 'TR' and (entry.id_top in match_types) and (entry.id_bottom in match_types) and (entry.id_left in match_types) and (entry.id_right not in match_types) : 
            tile_id = ndx
            break
        '''
        # Try to implement lanes by altering generation values.
        # A square tile is conceptually one unit wide no matter how many Blender units in size.
        # These values need to be in the range of -1.0 - 1.0.
        lane_right = -0.1
        lane_left = 0.1
        lane_center = 0.5
        lane_width = 0.25
        if layer.road_lane_type == '00': lane_value = 0.0
        if layer.road_lane_type == '01': lane_value = lane_right
        if layer.road_lane_type == '02': lane_value = lane_left
        if layer.road_lane_type == '03': 
            # Alternate odd right, even left.
            if float(float(index)/2.0) == int(float(index)/2.0):
                lane_value = lane_right
            else:
                lane_value = lane_left
        lane_end = 1.0
        lane_start = 0.0
        tangent_handles = []
        # Start off with both points in the center of the tile and apply deviations based upon the connection_type.
        x1 = lane_center
        y1 = lane_center
        x2 = lane_center
        y2 = lane_center	
        if connection_type == 'BR':			# Travel from bottom of tile to right.
            x1 = lane_center + lane_value
            y1 = lane_end
            x2 = lane_end
            y2 = lane_center + lane_value
            if x_dir == 1:
                x1 = lane_center - lane_value
                y1 = lane_start
                x2 = lane_end
                y2 = lane_center + lane_value
            elif y_dir == 1:
                x1 = lane_end
                y1 = lane_center - lane_value
                x2 = lane_center + lane_value
                y2 = lane_start

        elif connection_type == 'BL': 
            x1 = lane_center + lane_value
            y1 = lane_end
            x2 = lane_end
            y2 = lane_center + lane_value
            if x_dir == 1:
                x1 = lane_center - lane_value
                y1 = lane_start
                x2 = lane_end
                y2 = lane_center + lane_value
            if x_dir == -1:
                x1 = lane_center - lane_value
                y1 = lane_start
                x2 = lane_start
                y2 = lane_center - lane_value
            elif y_dir == 1:
                x1 = lane_start
                y1 = lane_center + lane_value
                x2 = lane_center + lane_value
                y2 = lane_start
                          
        elif connection_type == 'RT': 
            x1 = lane_center + lane_value
            y1 = lane_end
            y2 = lane_center + lane_value
            x2 = lane_end
            if y_dir == -1:
                x1 = lane_end
                y1 = lane_center - lane_value
                x2 = lane_center - lane_value
                y2 = lane_end
                
            
        elif connection_type == 'LT': 
            x1 = lane_start
            y1 = lane_center + lane_value
            y2 = lane_end
            x2 = lane_center - lane_value
            if x_dir == -1:
                x1 = lane_center + lane_value
                y1 = lane_end
                x2 = lane_start
                y2 = lane_center - lane_value

        # Vertical, horizontal.
        elif connection_type == '==': 
            x1 = lane_end 
            y1 = lane_center - lane_value
            x2 = lane_start
            y2 = lane_center - lane_value
            if x_dir == 1:
                x1 = lane_start 
                y1 = lane_center + lane_value
                x2 = lane_end
                y2 = lane_center + lane_value
        
        elif connection_type == '||': 
            x2 = lane_center - lane_value
            y2 = lane_end 
            x1 = lane_center - lane_value
            y1 = lane_start
            if y_dir == 1:
                x1 = lane_center + lane_value
                y1 = lane_end
                x2 = lane_center + lane_value
                y2 = lane_start
  
        # Single types.
        elif connection_type == 'ST':
            x1 = lane_center + lane_value
            y1 = lane_end
            x2 = lane_center + lane_value
            y2 = lane_center 
            if y_dir == -1:
                x1 = lane_center - lane_value
                y1 = lane_center
                x2 = lane_center - lane_value
                y2 = lane_end
                
        elif connection_type == 'SB': 
            x1 = lane_center - lane_value
            y1 = lane_start
            x2 = lane_center - lane_value
            y2 = lane_center
            if y_dir == 1:
                x1 = lane_center + lane_value
                y1 = lane_center
                x2 = lane_center + lane_value
                y2 = lane_start
            
        elif connection_type == 'SL': 
            y1 = lane_center + lane_value
            x1 = lane_start
            y2 = lane_center + lane_value
            x2 = lane_center
            if x_dir == -1:
                y2 = lane_center - lane_value
                x2 = lane_start
                y1 = lane_center - lane_value
                x1 = lane_center
            
            
        elif connection_type == 'SR': 
            x1 = lane_end
            y1 = lane_center - lane_value
            x2 = lane_center
            y2 = lane_center - lane_value
            if x_dir == 1:
                x1 = lane_center + lane_value
                y1 = lane_end
                x2 = lane_end
                y2 = lane_center + lane_value
        else:
            print("Unsupported connection type [%s] @(%ix%i)" % (connection_type, x,y))
        #print ("Direction of travel (%i, %i) along %s @(%ix%i)" % (x_dir, y_dir, connection_type,x,y))
        
        # Try to construct hi-lo Z-axis value based upon existing grid element name. (i.e. tile name)
        z_hi = 3.2
        z_lo = 1.2
        tile_name = collection[int(grid[x][y])].ob_name
        if (tile_name.find("high") != -1) or (tile_name.find("bridge") != -1):
            road_z1 = z_hi	#change to high road.
            road_z2 = z_hi	#change to high road.
        else:
            road_z1 = z_lo	#default to low road.
            road_z2 = z_lo	#default to low road.
        if (tile_name.find("ramp") != -1) or ((tile_name.find("hill") != -1) and (tile_name.find("bridge") == -1)):
            # Modify based upon ramp or hill and traveling direction.
            if (tile_name.find("RL") != -1):
                if x_dir == 1:
                    road_z1 = z_hi	
                    road_z2 = z_lo
                elif x_dir == -1:
                    road_z1 = z_lo
                    road_z2 = z_hi
            if (tile_name.find("LR") != -1):
                if x_dir == 1:
                    road_z1 = z_lo
                    road_z2 = z_hi
                elif x_dir == -1:
                    road_z1 = z_hi
                    road_z2 = z_lo
            if (tile_name.find("TB") != -1):
                if y_dir == 1:
                    road_z1 = z_lo
                    road_z2 = z_hi
                elif y_dir == -1:
                    road_z1 = z_hi
                    road_z2 = z_lo
            if (tile_name.find("BT") != -1):
                if y_dir == 1:
                    road_z1 = z_hi
                    road_z2 = z_lo
                elif y_dir == -1:
                    road_z1 = z_lo
                    road_z2 = z_hi
        
        coordinate_x = (unit_size * x) + (unit_size * x1) 	- (unit_size * (w/2))
        coordinate_y = (-unit_size * y) + (unit_size * y1)	- (-unit_size * (h/2))
        v1 = Vector ((coordinate_x - (unit_size/2), coordinate_y - (unit_size/2), road_z1))

        coordinate_x = (unit_size * x) + (unit_size * x2)	- (unit_size * (w/2))
        coordinate_y = (-unit_size * y) + (unit_size * y2)	- (-unit_size * (h/2))
        v2 = Vector ((coordinate_x - (unit_size/2),coordinate_y - (unit_size/2), road_z2))
        
        if layer.vehicle_path == False:
            s = "road_%s" % rig_name
            curve_name = returnCurveName(s,x,y)
            ob_curve_path = curveFromEndPoints(curve_name, v1, v2, curve_subdiv, noise, threshold)
            
            # Adjust tangent handles if curve is a diagonal. Try to form a rounded arc instead of a straight line.
            curve_types = ['BR', 'BL']
            if connection_type in curve_types:
                #print(connection_type,x_dir,y_dir)
                # Modifiy this curve's right and left handles.
                for spline in ob_curve_path.data.splines:
                    # There should only be one spline, we just created it.
                    p1 = spline.bezier_points[0]
                    p1.handle_right_type='FREE'
                    p1.handle_left_type='FREE'
                    co1 = p1.co
                    l = len(spline.bezier_points)
                    p2 = spline.bezier_points[(l-1)]
                    p2.handle_right_type='FREE'
                    p2.handle_left_type='FREE'
                    co2 = p2.co
                    if x_dir == 1:
                        p1.handle_left.x = co1.x
                        p1.handle_right.x = co1.x
                        p2.handle_left.y = co2.y
                        p2.handle_right.y = co2.y
                    elif x_dir == -1:
                        p1.handle_left.x = co1.x
                        p1.handle_right.x = co1.x
                        p2.handle_left.y = co2.y
                        p2.handle_right.y = co2.y
                    else:
                        p1.handle_left.y = co1.y
                        p1.handle_right.y = co1.y
                        p2.handle_left.x = co2.x
                        p2.handle_right.x = co2.x
                        
            curve_types = ['LT','RT']
            if connection_type in curve_types:
                #print(connection_type,x_dir,y_dir)
                # Modifiy this curve's right and left handles.
                for spline in ob_curve_path.data.splines:
                    # There should only be one spline, we just created it.
                    p1 = spline.bezier_points[0]
                    p1.handle_right_type='FREE'
                    p1.handle_left_type='FREE'
                    co1 = p1.co
                    l = len(spline.bezier_points)
                    p2 = spline.bezier_points[(l-1)]
                    p2.handle_right_type='FREE'
                    p2.handle_left_type='FREE'
                    co2 = p2.co
                    if y_dir == -1:
                        p1.handle_left.y = co1.y
                        p1.handle_right.y = co1.y
                        p2.handle_left.x = co2.x
                        p2.handle_right.x = co2.x
                    else:
                        p1.handle_left.x = co1.x
                        p1.handle_right.x = co1.x
                        p2.handle_left.y = co2.y
                        p2.handle_right.y = co2.y

            ob_curve_path.parent = mt_road_parts
            # Make sure this curve is linked to the scene.
            try:
                scene.objects.link(ob_curve_path)
            except:
                pass

        #curve_vector_set.append((v1,co1,co2))
        curve_vector_set.append(v1)
    curve_vector_set.append(v2)
 
    # Make a single curve that represents all curve parts.
    if layer.vehicle_path:
        points = []
        for v1 in curve_vector_set:
            points.append(v1)
        if len(points) > 0:
            # The trace produced points that we can use for constructing a new curve object.
            path_road_name = "path_road_%s" % rig_name
            ob_path_road = add_curve(scene, path_road_name, points)
            ob_path_road.parent = mt_road_parts
        
    # 5.) Add multiple follow path constraints to the rig. One for each curve element in the solution map.

    # For the next few lines of code we are going to override the User Preferences for keyframe interpolation type.
    keyInterp = bpy.context.user_preferences.edit.keyframe_new_interpolation_type
    keyHandle = bpy.context.user_preferences.edit.keyframe_new_handle_type
    bpy.context.user_preferences.edit.keyframe_new_interpolation_type ='LINEAR'
    bpy.context.user_preferences.edit.keyframe_new_handle_type ='VECTOR'

    ob = bpy.data.objects.get(rig_name)
    if ob != None:
        ob.constraints.clear()
        ad = ob.animation_data
        if ad != None:
           ob.animation_data_clear()
        end_frame = 1
        rnd = (unit_size/2) - int(random.random() * unit_size)
        frame_delta = layer.vehicle_speed + rnd
        l = len(solution_coordinates)
        if layer.vehicle_path:
            # Single path solution for vehicles. (less accurate, more convienent)
            #solution_coordinates = [(0,0)]
            #for i,(x,y) in enumerate(solution_coordinates):
            #v = 1.0/len(solution_coordinates)
            start_frame = end_frame
            end_frame = l * frame_delta
            if ob_path_road:
                eraseAllKeyframes(ob_path_road)
                # Assign a follow path constraint the rig.
                cn = returnFollowPathConstraint(path_road_name, ob, ob_path_road, 1.0)
                if cn != None:
                    # Insert two linear keyframes to animate along the path.
                    cn.offset_factor = 0.0
                    cn.keyframe_insert(data_path='offset_factor', frame=(int(start_frame)))
                    cn.offset_factor = 0.98
                    cn.keyframe_insert(data_path='offset_factor', frame=(int(end_frame)))
            
                # Create a steering Empty for this path.
                mt_steering_name = returnIndexedName(("mt_steer_%s" % rig_name), c)
                mt_steering = bpy.data.objects.get(mt_steering_name)
                if mt_steering == None:
                    # Does not exist, create.
                    mt_steering = bpy.data.objects.new(mt_steering_name, None)
                try:
                    scene.objects.link(mt_steering)
                except:
                    pass
                mt_steering.parent = mt_road_parts
                eraseAllKeyframes(mt_steering)
                cn = returnFollowPathConstraint(mt_steering_name, mt_steering, ob_path_road, 1.0)
                if cn != None:
                    # Insert two linear keyframes to animate along the path.
                    cn.offset_factor = 0.02
                    cn.keyframe_insert(data_path='offset_factor', frame=(int(start_frame)))
                    cn.offset_factor = 1.0
                    cn.keyframe_insert(data_path='offset_factor', frame=(int(end_frame)))
                pb = ob.pose.bones.get("Wheel Main Steer")		#HARD CODED reference to a bone name in the vehicle rig.
                if pb != None:
                    # We assume this pose bone has a follow path constraint already installed.
                    pb.constraints[0].target = mt_steering
        else:
            # Multiple curves solution for vehicles. (more accurate, less convienent)
            for i,(x,y) in enumerate(solution_coordinates):
                start_frame = end_frame
                end_frame = start_frame + frame_delta
                s = "road_%s" % rig_name			# same name as above loop.
                curve_name = returnCurveName(s,x,y)
                ob_curve = bpy.data.objects.get(curve_name)
                if ob_curve != None:
                    eraseAllKeyframes(ob_curve)
                    # Assign a follow path constraint the rig.
                    cn = returnFollowPathConstraint(curve_name, ob, ob_curve, 0.0)
                    if cn != None:
                        # Insert two linear keyframes to animate along the path.
                        cn.offset_factor = 0.0
                        cn.keyframe_insert(data_path='offset_factor', frame=(int(start_frame)))
                        cn.offset_factor = 1.0
                        cn.keyframe_insert(data_path='offset_factor', frame=(int(end_frame)))

                        # Insert keyframes for influence based upon keyrange.
                        cn.influence = 0.0
                        cn.keyframe_insert(data_path='influence', frame=(int(start_frame-1)))
                        cn.influence = 1.0
                        cn.keyframe_insert(data_path='influence', frame=(int(start_frame)))
                        cn.influence = 1.0
                        cn.keyframe_insert(data_path='influence', frame=(int(end_frame-1)))
                        if i != (l-1):
                            cn.influence = 0.0
                        cn.keyframe_insert(data_path='influence', frame=(int(end_frame)))
                        
                        # Find the fcurves we just created (there has got to be a better way..?)
                        action = returnActionFromObject(ob)
                        if action != None:
                            for fcurve in action.fcurves:
                                if fcurve.data_path.find(cn.name) != -1:
                                    # This curve belongs to the current constraint.
                                    if (fcurve.data_path.find('offset_factor') != -1) or (fcurve.data_path.find('influence') != -1):
                                        #Change the fcurve keyframes to linear with vector handles.
                                        for keyframe in fcurve.keyframe_points:
                                            keyframe.handle_left_type = 'VECTOR'
                                            keyframe.handle_right_type = 'VECTOR'
                                            keyframe.interpolation = 'LINEAR'
                                            keyframe.co.x = int(keyframe.co.x)  # Make sure handles fall on INT only keyframes too.
                    else:
                        print("Failed to create or fetch constraint [%s]." % (curve_name))
                else:
                    print("Failed to create or fetch object [%s]." % (curve_name))

    # Restore User Preferences to what it was.
    bpy.context.user_preferences.edit.keyframe_new_interpolation_type = keyInterp
    bpy.context.user_preferences.edit.keyframe_new_handle_type = keyHandle

    # 6.) Setup drivers to make the wheels rotate based upon the distance traveled along the path and the wheel diameter.
    bone_base = bpy.data.objects[rig_name].data.bones.get("Base")
    bone_base["wheel_diameter"] = 33.0	#HARD CODED change to your wheel size.

    # Add a driver to each wheel.
    wheel_driver_list = ['Wheel_Rear.R','Wheel_Rear.L','Wheel_front.R','Wheel_front.L']
    for wheel_name in wheel_driver_list:
        wheel_driver_bone = bpy.data.objects[rig_name].pose.bones.get(wheel_name)
        if wheel_driver_bone != None:
            driver = wheel_driver_bone.driver_remove('rotation_euler',-1)
            driver = wheel_driver_bone.driver_add('rotation_euler',0).driver
            driver.type = 'SCRIPTED'
            var = driver.variables.new()
            var.targets[0].id_type = 'SCENE'
            var.targets[0].id =  bpy.context.scene.id_data
            var.targets[0].data_path = "frame_current" 
            var.name = "this_frame" 
            driver.expression =  'returnWheelRotationFromDistance("%s", "%s", %s, this_frame)' % (rig_name, wheel_name, bone_base["wheel_diameter"])    #((360.0/(8.0*1.0))-bpy.data.objects["rig_car"].data.bones["Base"]["distance"]'    # distance assigned to wheel rotation.
        else:
            print ("wheel driver bone [%s] not found." % wheel_name)
            
    #Adjust the wheel base for bones.
    wheel_base_list = ['Wheel_Rear.R','Wheel_Rear.L','Wheel_track.R','Wheel_track.L']
    for wheel_name in wheel_base_list:
        wheel_base_bone = bpy.data.objects[rig_name].pose.bones.get(wheel_name)
        if wheel_base_bone != None:
            # Adjust the wheel base distance here.
            if wheel_name.find("Rear") != -1:
                wheel_base_bone.location.y = wheel_base_rear
                if wheel_name.find(".L") != -1:
                    wheel_base_bone.location.x = wheel_base_left
                else:
                    wheel_base_bone.location.x = wheel_base_right
            else:
                # The track bone needs alternate logic, inverted for some reason...? (pose space?)
                if wheel_name.find(".L") != -1:
                    wheel_base_bone.location.x = -wheel_base_left
                else:
                    wheel_base_bone.location.x = -wheel_base_right
                wheel_base_bone.location.z = wheel_base_front + 0.05 
        else:
            print ("wheel base bone [%s] not found." % wheel_name)
                
def register():
    # registering name space won't work at this time because of flaky _RestrictedData concept that developers through in to just keep Blender from crashing on startup.
    #print ("registering auto drive name space.")
    #bpy.app.driver_namespace["returnWheelRotationFromDistance"] = returnWheelRotationFromDistance
    #print(bpy.app.driver_namespace["returnWheelRotationFromDistance"])
    #print(returnWheelRotationFromDistance("rig_car", "Wheel_front.R", 33.0, 1))
    pass
    
def unregister():
    pass
