# AddOn Tile Proc (c) 2014 Atom

import bpy, random, math
import os, sys

from operator import attrgetter

from .util import to_console
from .util import returnNameForNumber
from .util import returnTileName
from .util import returnItemName
from .util import returnRandomColor
from .util import returnRandomGauss
from .util import removeObjectFromMemory
from .util import returnTileNeighborTypes
from .util import returnMapNeighborTypes
from .util import returnSelfDeterminedByNeighbors

from .util import OBJECT_PREFIX
from .util import ENTRY_NAME
from .util import ITEM_NAME
from .util import MAX_NAME_SIZE

from .util import isBusy

from .auto_drive import generateDrivingVehicle
from .auto_drive import generateFlyingBird
    
############################################################################
# Operator code.
############################################################################
# Create operator to rename this object with the tileproc_ preifix.   
class OBJECT_OT_generate_from_tiles(bpy.types.Operator):
    bl_label = "Generate From Tiles"
    bl_idname = "op.generate_from_tiles"
    bl_description = "Click this button to generate a tile based landscape."
                         
    def returnBlankGrid(self, width, height, value = None):
        grid = [[0 for y in range(height)] for x in range(width)]
        for x in range(0,width):
            for y in range(0,height):
                if value == None:
                    grid[x][y] = int(-1)        # Blank out the grid.
                else:
                    grid[x][y] = value
        return grid
        
    def printGrid (self, grid, title):
        w = len(grid)
        h = len(grid[0])
        print("\n%s" % title)
        for x in range(0,w):
            s = ""
            for y in range(0,h):
                s += str(grid[x][y])
            print(s)
    
    def invertMap(self, grid):
        w = len(grid)
        h = len(grid[0])
        inverted_grid = self.returnBlankGrid(w,h)
        for x in range(0,w):
            for y in range(0,h):
                if grid[x][y] == '.':
                    inverted_grid[x][y] = '#'
                else:
                    inverted_grid[x][y] = '.'
        return inverted_grid
            
    def returnRndAreaMap(self, width, height, area, offset_x, offset_y, rnd_seed):
        random.seed(rnd_seed)
        grid = self.returnBlankGrid(width,height,'#')
        # Determine are from the center.
        x_length = int((width * area) / 1.0)
        y_length = int((height * area) / 1.0)
        x_len_div2 = int(x_length/2.0)
        y_len_div2 = int(y_length/2.0)
        x_center = int(width/2.0) + int((offset_x * x_length))
        y_center = int(height/2.0) + int((offset_y * y_length))
        # Generate fill range with offset.
        xstart = x_center - x_len_div2
        ystart = y_center - y_len_div2
        xend = xstart + x_length
        yend = ystart + y_length
        # Check bounds.
        if xstart < 0: xstart = 0
        if xstart > width: xstart = width
        if ystart < 0: ystart = 0
        if ystart > height: ystart = height
        if xend < 0: xend = 0
        if xend > width: xend = width
        if yend < 0: yend = 0
        if yend > height: yend = height
        coordinates = [(i, j) for i in range(xstart,xend) for j in range(ystart,yend)]
        l = len(coordinates)
        if l > 0:
            random.shuffle(coordinates)
            #random.shuffle(coordinates, random.random)
            #r = random.gauss(5,1)
            #random.shuffle(coordinates, lambda: r)
            i = 0
            for (x,y) in coordinates:
                percent_so_far = i / l
                if percent_so_far < area:
                    # Have not exceeded area, add empty space.
                    grid[x][y] = '.'
                i +=1
        return grid
   
    def returnMazeMap(self, width, height, branch_rate, mask_map, rnd_seed):
        random.seed(rnd_seed)
        xwide = width
        yhigh = height
        #each cell of the maze is one of the following:
        # '#' is wall
        # '.' is empty space
        # ',' is exposed but undetermined
        # '?' is unexposed and undetermined
        field = self.returnBlankGrid(width,height)
 
        # Blank out the grid, maze style (i.e. with ?, NOT -1).
        grid = self.returnBlankGrid(width,height,'?')
                
        if mask_map == None:
            # Randomly add some pre-walls to see if roads can go around them.
            # The more pre-walls, the more "open-space" in the final generation.
            l = int(((width+height)/2)/8)
            size = random.randint(1,l)  #Try to make road block related to overall grid size.
            xrnd = random.randint(0, xwide-size)
            yrnd = random.randint(0, yhigh-size)
            for x in range(xrnd,(xwide-size)):
                for y in range(yrnd,(yhigh-size)):
                    grid[x][y] = '#'                        # Discovered solid wall character.
        else:
            # Use passed map as a mask which is assumed to be the same size as width and height.
            for x in range(xwide):
                for y in range(yhigh):
                    if mask_map[x][y] == '#':
                        # Avoid generating into this grid space.
                        grid[x][y] = '#'                    # Discovered solid wall character.
            
        # Convert grid to field.
        for x in range(xwide):
            for y in range(yhigh):
                field[x][y] = grid[x][y]
        #self.printGrid(field,"mask map")

        #list of coordinates of exposed but undetermined cells.
        frontier = []

        def carve(x, y):
            '''Make the cell at x,y a space.
            
            Update the frontier and field accordingly.
            Note: this does not remove the current cell from frontier, it only adds new cells.
            '''
            extra = []
            field[x][y] = '.'
            if x > 0:
                if field[x-1][y] == '?':
                    field[x-1][y] = ','
                    extra.append((x-1,y))
            if x < xwide - 1:
                if field[x+1][y] == '?':
                    field[x+1][y] = ','
                    extra.append((x+1,y))
            if y > 0:
                if field[x][y-1] == '?':
                    field[x][y-1] = ','
                    extra.append((x,y-1))
            if y < yhigh - 1:
                if field[x][y+1] == '?':
                    field[x][y+1] = ','
                    extra.append((x,y+1))
            random.shuffle(extra)
            frontier.extend(extra)

        def harden(x, y):
            field[x][y] = '#'   #Make the cell at x,y a wall.

        def check(x, y, nodiagonals = True):
            '''Test the cell at x,y: can this cell become a space?
            
            True indicates it should become a space,
            False indicates it should become a wall.
            '''
            edgestate = 0
            if x > 0:
                if field[x-1][y] == '.':
                    edgestate += 1
            if x < xwide-1:
                if field[x+1][y] == '.':
                    edgestate += 2
            if y > 0:
                if field[x][y-1] == '.':
                    edgestate += 4
            if y < yhigh-1:
                if field[x][y+1] == '.':
                    edgestate += 8
            
            if nodiagonals:
                #if this would make a diagonal connection, forbid it
                #the following steps make the test a bit more complicated and are not necessary,
                #but without them the mazes don't look as good
                if edgestate == 1:
                    if x < xwide-1:
                        if y > 0:
                            if field[x+1][y-1] == '.':
                                return False
                        if y < yhigh-1:
                            if field[x+1][y+1] == '.':
                                return False
                    return True
                elif edgestate == 2:
                    if x > 0:
                        if y > 0:
                            if field[x-1][y-1] == '.':
                                return False
                        if y < yhigh-1:
                            if field[x-1][y+1] == '.':
                                return False
                    return True
                elif edgestate == 4:
                    if y < yhigh-1:
                        if x > 0:
                            if field[x-1][y+1] == '.':
                                return False
                        if x < xwide-1:
                            if field[x+1][y+1] == '.':
                                return False
                    return True
                elif edgestate == 8:
                    if y > 0:
                        if x > 0:
                            if field[x-1][y-1] == '.':
                                return False
                        if x < xwide-1:
                            if field[x+1][y-1] == '.':
                                return False
                    return True
                return False
            else:
                #diagonal walls are permitted
                if [1,2,4,8].count(edgestate):
                    return True
                return False
            
        # Choose a original point at random and carve it out.
        xchoice = random.randint(0, xwide-1)
        ychoice = random.randint(0, yhigh-1)
        
        # Try to avoid starting inside a wall.
        c = 0
        while field[xchoice][ychoice] != '?':       
            xchoice = random.randint(0, xwide-1)
            ychoice = random.randint(0, yhigh-1)
            c += 1
            if c > 512: break
        #print ("Carve decided to start at (%ix%i) after %i tries." % (xchoice,ychoice,c))
        #self.printGrid(field,"Maze Field")
        carve(xchoice,ychoice)

        from math import e

        while(len(frontier)):
            #select a random edge
            pos = random.random()
            pos = pos**(e**-branch_rate)
            choice = frontier[int(pos*len(frontier))]
            if check(*choice):
                carve(*choice)
            else:
                harden(*choice)
            frontier.remove(choice)
            
        #set unexposed cells to be walls
        for x in range(xwide):
            for y in range(yhigh):
                if field[x][y] == '?':
                    field[x][y] = '#'
        return field

    def joinMeshes(self, hostMesh, additionMesh):
        # Thanks to Liero for this code!
        #http://blenderartists.org/forum/showthread.php?286135-How-To-Add-The-Mesh-From-One-Object-To-Another
        sel = [hostMesh, additionMesh]
        mes = bmesh.new()
        
        for me in sel:
            tmp = bmesh.new()
            tmp.from_mesh(me)
            num = len(mes.verts)
        
            for nv in tmp.verts:
                mes.verts.new(nv.co.to_tuple())
            mes.verts.index_update()
            for nf in tmp.faces:
                vvv = [num + v.index for v in nf.verts]
                bm_face = mes.faces.new([mes.verts[v] for v in vvv])
                bm_face.material_index = nf.material_index
            tmp.free()

        mes.to_mesh(hostMesh)
        
    def itemsToTile (self, scene, ob_parent, layer, area_element, x, y, grid, collection, rnd_seed):
        #random.seed(rnd_seed)
        w = len(grid)
        h = len(grid[0])
        if layer.name == 'Trees':
            candidates_list = self.returnTreeList(collection)
        if layer.name == 'Bushes':
            candidates_list = self.returnBushList(collection)
        if layer.name == 'Flowers':
            candidates_list = self.returnFlowerList(collection)
        if layer.name == 'Street Assets':
            candidates_list = self.returnStreetAssetList(collection)
        if layer.name == 'Advertising Assets':
            candidates_list = self.returnAdvertisingAssetList(collection)
        l = len(candidates_list)
        if l > 0:
            for c in range(layer.item_count):
                if area_element == '.':
                    entry = collection[grid[x][y]]
                    should_skip = False
                    if random.random() > layer.item_coverage: should_skip = True
                    if (entry.ob_name.find("river") != -1): should_skip = True
                    if should_skip == False:    # Skip some sometimes.
                        ndx = int(random.random() * l)
                        ob_name = collection[candidates_list[ndx]].ob_name
                        ob_candidate = bpy.data.objects.get(ob_name)
                        if ob_candidate != None:
                            ob_candidates = []
                            ob_candidates.append(ob_candidate)
                            for d,ob_clone_source in enumerate(ob_candidates):
                                n = "%s-%s_%s" % (ITEM_NAME, returnNameForNumber(d), layer.name)
                                item_name = returnItemName(n, x, y, c) # Name of item to be created.
                                temp_ob_item = bpy.data.objects.get(item_name)
                                if temp_ob_item == None:
                                    # If it does not exist.
                                    if ob_candidate.type == 'EMPTY':
                                        # Make a copy.
                                        temp_ob_item = ob_clone_source.copy()
                                        temp_ob_item.name = item_name
                                    else:
                                        # Create it with new datablock so this clone is an instance.
                                        temp_ob_item = bpy.data.objects.new(item_name,ob_clone_source.data)     #ob_clone_source.copy()
                                else:
                                    if ob_candidate.type == 'EMPTY':
                                        # Empties do not have a datablock, skip.
                                        pass
                                    else:
                                        # It already exists, assign it a new datablock.
                                        temp_ob_item.data = ob_clone_source.data
                                try:
                                    scene.objects.link(temp_ob_item)
                                except:
                                    pass				# Already linked?
         
                                # Generate a random location within this tile's area which is 8x8 Blender units in size.
                                 #Fetch information from the AddOn preferences.
                                unit_size = bpy.context.user_preferences.addons['tile_proc'].preferences.unit_size
                                x_center = unit_size * x - (unit_size * (w/2))
                                y_center = -unit_size * y - (-unit_size * (h/2))
                                area_size = unit_size * layer.area  #layer.item_coverage
                                x_offset = (random.random() * area_size)
                                y_offset = (random.random() * area_size)
                                temp_ob_item.location.x = x_center + (x_offset-(area_size/2))
                                temp_ob_item.location.y = y_center + (y_offset-(area_size/2))
                                
                                # Adjust Z-location based upon tile type in this grid element.
                                z_pos = 0.96
                                if (entry.ob_name.find("dune") != -1) or (entry.ob_name.find("high") != -1) or (entry.ob_name.find("bridge") != -1):
                                    
                                    if (entry.ob_name.find("low") == -1):
                                        # Hills have higher z postion.
                                        z_pos = 2.96
                                    if (entry.ob_name.find("end") != -1):
                                        print(entry.ob_name)
                                        if (entry.ob_name.find("T") != -1) or (entry.ob_name.find("B") != -1) or (entry.ob_name.find("R") != -1) or (entry.ob_name.find("L") != -1):
                                            z_pos = 0.96
                                #if (entry.ob_name.find("hill") != -1) or (entry.ob_name.find("ramp") != -1):
                                if (entry.ob_name.find("ramp") != -1):
                                    if (entry.ob_name.find("low") == -1):
                                        # Modify for ramps.
                                        z_pos = 1.96

                                temp_ob_item.location.z = z_pos
                                rnd_z = 0
                                rnd_size = (random.random() * 2) * layer.item_rnd_size
                                temp_ob_item.scale = (1.0+rnd_size,1.0+rnd_size,1.0+rnd_size)
                                if layer.name == 'Street Assets':
                                    if entry.ob_name.find("L") != -1: rnd_z = 180
                                    elif entry.ob_name.find("R") != -1: rnd_z = 0
                                    elif entry.ob_name.find("T") != -1: rnd_z = 90
                                    elif entry.ob_name.find("B") != -1: rnd_z = 270
                                    else:
                                        print ("Unsupported alignment [%s]" % entry.ob_name)
                                    if entry.ob_name.find("LR") != -1: rnd_z = 90
                                    if entry.ob_name.find("RL") != -1: rnd_z = 90
                                    if entry.ob_name.find("TB") != -1: rnd_z = 0
                                    if entry.ob_name.find("BL") != -1: rnd_z = 180 + 45
                                    if entry.ob_name.find("LT") != -1: rnd_z = 90 + 45
                                    if entry.ob_name.find("TL") != -1: rnd_z = 90 + 45
                                    if entry.ob_name.find("TR") != -1: rnd_z = 45
                                    if entry.ob_name.find("RT") != -1: rnd_z = 45
                                    if entry.ob_name.find("BT") != -1: rnd_z = 180
                                    if entry.ob_name.find("RB") != -1: rnd_z = 315
                                    if entry.ob_name.find("LB") != -1: rnd_z = 180 + 45
                                    
                                    # Assign tile space based upon tile name, mainly looking for L R T B combinations.
                                    x_offset = 0.0
                                    y_offset = 0.0
                                    if entry.ob_name.find("L") != -1: 
                                        x_offset = 0.8
                                        y_offset = 0.49
                                    if entry.ob_name.find("R") != -1: 
                                        x_offset = 0.2
                                        y_offset = 0.49
                                    if entry.ob_name.find("T") != -1: 
                                        x_offset = 0.49
                                        y_offset = 0.2
                                    if entry.ob_name.find("B") != -1: 
                                        x_offset = 0.49
                                        y_offset = 0.8
                                    if entry.ob_name.find("LR") != -1:
                                        x_offset = 0.49
                                        y_offset = 0.2
                                    if entry.ob_name.find("BL") != -1:
                                        x_offset = 0.57
                                        y_offset = 0.57
                                    if entry.ob_name.find("LT") != -1:
                                        x_offset = 0.6
                                        y_offset = 0.46
                                    if entry.ob_name.find("TL") != -1:
                                        x_offset = 0.6
                                        y_offset = 0.46
                                    if entry.ob_name.find("TR") != -1:
                                        x_offset = 0.43
                                        y_offset = 0.43
                                    if entry.ob_name.find("RT") != -1:
                                        x_offset = 0.43
                                        y_offset = 0.43
                                    if entry.ob_name.find("TB") != -1:
                                        x_offset = 0.2
                                        y_offset = 0.49
                                    if entry.ob_name.find("BT") != -1:
                                        x_offset = 0.8
                                        y_offset = 0.49
                                    if entry.ob_name.find("RB") != -1:
                                        x_offset = 0.45
                                        y_offset = 0.58
                                    if entry.ob_name.find("LB") != -1:
                                        x_offset = 0.57
                                        y_offset = 0.57
                                    if entry.ob_name.find("RL") != -1:
                                        x_offset = 0.49
                                        y_offset = 0.2
                                    area_size = unit_size

                                    #temp_ob_item.location.x = ob_clone_source.location.x - x_center + ((x_offset * unit_size)-(area_size/2))
                                    #temp_ob_item.location.y = ob_clone_source.location.y - y_center + ((y_offset * unit_size)-(area_size/2))
                                    
                                    temp_ob_item.location.x = x_center + ((x_offset * unit_size)-(area_size/2))
                                    temp_ob_item.location.y = y_center + ((y_offset * unit_size)-(area_size/2))
                                        
                                else:
                                    # Randomize the Z-rotation.
                                    rnd_z = random.random() * 360   # Degrees in a circle.
                                temp_ob_item.rotation_euler.z = math.radians(rnd_z)
                                try:
                                    temp_ob_item.parent = ob_parent
                                except:
                                    pass				# Already parented?
                        else:
                            print("Object [%s] is in the candidate list but does not exist in bpy.data?" % ob_name)
                    else:
                        #print ("should skip to avoid or because random")
                        pass
                else:
                    #print ("Not a buildable space.")
                    for d in range(8):                                     # HARD CODE guessing that we won't have more than 8 items in a group or erase cleanup won't work.
                        n = "%s-%s_%s" % (ITEM_NAME, returnNameForNumber(d), layer.name)
                        item_name = returnItemName(n, x, y, c)              # Name of item to be created.
                        temp_ob_item = bpy.data.objects.get(item_name)
                        if temp_ob_item != None:
                            try:
                                scene.objects.unlink(temp_ob_item)          # Unlink from scene if in scene.
                            except:
                                # Already unlinked?
                                pass
                                #print ("Error unlinking [%s], already unlinked?" % item_name)
                        pass
        else:
            print ("No source mesh candidates discovered for %s." % layer.name)
                        
    
    def gridToObjects (self, scene, ob_parent, grid, collection):
        # UPDATE OBJECTS IN SCENE.
        # Loop through the grid and make a tile object for each position in the grid.
        w = len(grid)
        h = len(grid[0])
        for x in range(0,w):
            for y in range(0,h):
                #to_console ("Working on %ix%i." % (x,y))
                n = "%s_%s" % (ENTRY_NAME, ob_parent.name[-6:])
                tile_name = returnTileName(n, x, y) # Name of tile to be created.
                temp_ob_tile = bpy.data.objects.get(tile_name)
                if temp_ob_tile != None:
                    try:
                        scene.objects.unlink(temp_ob_tile)      # Unlink from scene if in scene.
                    except:
                        # Already unlinked?
                        pass
                        #to_console ("Error unlinking [%s], already unlinked?" % tile_name)
                        
                if grid[x][y] != -1:
                    if grid[x][y] == -2:
                        # This grid is occupied by a 2x asset.
                        pass
                    else:
                        # This grid is occupied by 1x tile, link it to the scene.
                        ob_name = collection[grid[x][y]].ob_name        # Name of source object where new tile will get it's mesh data.
                        ob_candidate = bpy.data.objects.get(ob_name)
                        if ob_candidate != None:
                            if temp_ob_tile == None:
                                # If it does not exist, create it with new datablock.
                                temp_ob_tile = bpy.data.objects.new(tile_name,ob_candidate.data)
                            else:
                                # It already exists, assign it a new datablock.
                                temp_ob_tile.data = ob_candidate.data
                            try:
                                scene.objects.link(temp_ob_tile)
                            except:
                                # Already linked?
                                pass
                                #to_console ("Error linking [%s], already linked?" % tile_name)
                                
                            # In our case, the tiles we are using from http://www.blendswap.com/blends/view/75021 are 8x8 in size.
                            unit_size = bpy.context.user_preferences.addons['tile_proc'].preferences.unit_size
                            if collection[grid[x][y]].is_2x_asset:
                                temp_ob_tile.location.x = unit_size * x - (unit_size * (w/2)) + (unit_size/2)
                                temp_ob_tile.location.y = -unit_size * y - (-unit_size * (h/2)) - (unit_size/2)
                                temp_ob_tile.location.z = 0.0
                            else:
                                temp_ob_tile.location.x = unit_size * x - (unit_size * (w/2))
                                temp_ob_tile.location.y = -unit_size * y - (-unit_size * (h/2))
                            
                            try:
                                temp_ob_tile.parent = ob_parent
                            except:
                                # Already parented?
                                pass
                                #to_console ("Error parenting [%s], already parented?" % tile_name)
                        else:
                            to_console("Object [%s] is in the tile list but does not exist in bpy.data?" % ob_name)
       
    def returnTileOrientationViaType(self, id_type, tile_index, collection):
        result = None
        tile = collection[tile_index]
        
        # A Quad intersection.
        if tile.id_top == id_type:
            if tile.id_top == tile.id_bottom and tile.id_top == tile.id_left and tile.id_top == tile.id_right: return "XX"

        # T intersections.
        if tile.id_bottom == id_type:
            if (tile.id_bottom == tile.id_right) and (tile.id_bottom == tile.id_left): return "TT"
        if tile.id_top == id_type:
            if (tile.id_top == tile.id_right) and (tile.id_top == tile.id_left): return "TB"
        if tile.id_top == id_type:
            if (tile.id_top == tile.id_bottom) and (tile.id_top == tile.id_right): return "TL"
        if tile.id_top == id_type:
            if (tile.id_top == tile.id_bottom) and (tile.id_top == tile.id_left): return "TR"
        
        # Vertical, horizontal.
        if tile.id_top == id_type:
            if tile.id_top == tile.id_bottom: return "||"
        if tile.id_left == id_type:
            if tile.id_left == tile.id_right: return "=="                       
        
        # Corner types.
        if tile.id_top == id_type:
            if tile.id_top == tile.id_right: return "RT"
        if tile.id_top == id_type:
            if tile.id_top == tile.id_left: return "LT"
        if tile.id_bottom == id_type:
            if tile.id_bottom == tile.id_left: return "BL"
        if tile.id_bottom == id_type:
            if tile.id_bottom == tile.id_right: return "BR"
                
        '''
        # Single types.
        if top == "." and bottom == "#" and left == "#" and right == "#": result = "ST"
        if top == "#" and bottom == "." and left == "#" and right == "#": result = "SB"
        if top == "#" and bottom == "#" and left == "." and right == "#": result = "SL"
        if top == "#" and bottom == "#" and left == "#" and right == ".": result = "SR"
        '''
        return result
 
    def modifyVerticalStrips(self, hill_threshold, area_map, grid, candidates, connectors, collection):
        # NOTE: grid and map are assumed to be of the same dimensions at this time.
        lst_strips = self.scanForVertical (3,hill_threshold,area_map)   
        l_hrc = len(candidates)-1
        for (x, strip_start, strip_end) in lst_strips:
            if (strip_end-strip_start) > 1:
                # A short hill.
                grid[x][(strip_start+1)] = connectors[0]
                grid[x][(strip_end-1)] = connectors[1]
                new_strip_start = strip_start + 2
                new_strip_end = strip_end - 1
                if (new_strip_end-new_strip_start) > 0:
                    for i in range(new_strip_start, new_strip_end):
                        ndx = int(random.random() * l_hrc)
                        grid[x][i] = candidates[ndx]

    def modifyHorizontalStrips(self, hill_threshold, area_map, grid, candidates, connectors, collection):
        # NOTE: grid and map are assumed to be of the same dimensions at this time.
        lst_strips = self.scanForHorizontal (3,hill_threshold,area_map)
        l_hrc = len(candidates)-1
        for (y, strip_start, strip_end) in lst_strips:
            if (strip_end-strip_start) > 1:
                # A short hill.
                grid[(strip_start+1)][y] = connectors[0]
                grid[(strip_end-1)][y] = connectors[1]
                new_strip_start = strip_start + 2
                new_strip_end = strip_end - 1
                if (new_strip_end-new_strip_start) > 0:
                    for i in range(new_strip_start, new_strip_end):
                        ndx = int(random.random() * l_hrc)
                        grid[i][y] = candidates[ndx] 
      
    def returnHighRoadsVerticalCandidates(self, collection):
        candidates = []
        for (i,entry) in enumerate(collection):
            if self.isVerticalHighRoad(i,collection):
                candidates.append(i)
        return candidates

    def returnHighRoadsHorizontalCandidates(self, collection):
        candidates = []
        for (i,entry) in enumerate(collection):
            if self.isHorizontalHighRoad(i,collection):
                candidates.append(i)
        return candidates
        
    def returnHighCanalsVerticalCandidates(self, collection):
        candidates = []
        for (i,entry) in enumerate(collection):
            if self.isVerticalHighCanal(i,collection):
                candidates.append(i)
        return candidates

    def returnHighCanalsHorizontalCandidates(self, collection):
        candidates = []
        for (i,entry) in enumerate(collection):
            if self.isHorizontalHighCanal(i,collection):
                candidates.append(i)
        return candidates

    def returnHighRiversVerticalCandidates(self, collection):
        candidates = []
        for (i,entry) in enumerate(collection):
            if self.isVerticalHighRiver(i,collection):
                candidates.append(i)
        return candidates

    def returnHighRiversHorizontalCandidates(self, collection):
        candidates = []
        for (i,entry) in enumerate(collection):
            if self.isHorizontalHighRiver(i,collection):
                candidates.append(i)
        return candidates
                
    def scanForVertical (self, strip_length, hill_threshold, grid):
        w = len(grid)
        h = len(grid[0])
        lst_strip_sets = []
        # Only consider strips if they are not along the edge of the grid.
        for x in range(1,(w-1)):
            strip_start = -1
            strip_end = -1
            for y in range(1,(h-1)):
                if strip_start == -1:
                    # We have not started yet.
                    if grid[x][y] == ".":
                        # Detected the ASCII symbol for a passage. Time to start.
                        strip_start = y
                else:
                    if grid[x][y] == ".":
                        # Detected the ASCII symbol for a passage.
                        # Horizontal neighbors must be walls.
                        if (grid[x+1][y] == "#") and (grid[x-1][y] == "#"):
                            # Extend the length of this stip.
                            strip_end = y
                        else:
                            # A break in the current vertical strip.
                            if strip_start != -1:
                                strip_end = y-1
                                # A strip was started in this column, let's review it's length.
                                if (strip_end-strip_start) >= strip_length:
                                    # Add this range as a candidate for high_low modification (i.e. ramps, bridges).
                                    if random.random() < hill_threshold:
                                        lst_strip_sets.append((x, strip_start, strip_end)) 
                            # Reset the strip variables so we can detect another strip.
                            strip_start = -1
                            strip_end = -1
                    else:
                        # Other symbols indicate a break in the current vertical strip.
                        if strip_start != -1:
                            strip_end = y-1
                            # A strip was started in this column, let's review it's length.
                            if (strip_end-strip_start) >= strip_length:
                                # Add this range as a candidate for high_low modification (i.e. ramps, bridges).
                                if random.random() < hill_threshold:
                                    lst_strip_sets.append((x, strip_start, strip_end)) 
                        # Reset the strip variables so we can detect another strip.
                        strip_start = -1
                        strip_end = -1
            # Check to see if a strip was in progress when we reached the edge.
            if strip_start != -1:
                strip_end = y-1
                # A strip was started in this column, let's review it's length.
                if (strip_end-strip_start) >= strip_length:
                    # Add this range as a candidate for high_low modification (i.e. ramps, bridges).
                    if random.random() < hill_threshold:
                        lst_strip_sets.append((x, strip_start, strip_end)) 
        return lst_strip_sets
        
    def scanForHorizontal (self, strip_length, hill_threshold, grid):
        w = len(grid)
        h = len(grid[0])
        lst_strip_sets = []
        # Only consider strips if they are not along the edge of the grid.
        for y in range(1,(h-1)):
            strip_start = -1
            strip_end = -1
            for x in range(1,(w-1)):
                if strip_start == -1:
                    # We have not started yet.
                    if grid[x][y] == ".":
                        # Detected the ASCII symbol for a passage. Time to start.
                        strip_start = x
                else:
                    if grid[x][y] == ".":
                        # Detected the ASCII symbol for a passage.
                        # Vertical neighbors must be walls.
                        if (grid[x][y+1] == "#") and (grid[x][y-1] == "#"):
                            # Extend the length of this stip.
                            strip_end = x
                        else:
                            # A break in the current vertical strip.
                            if strip_start != -1:
                                strip_end = x-1
                                # A strip was started in this column, let's review it's length.
                                if (strip_end-strip_start) >= strip_length:
                                    # Add this range as a candidate for high_low modification (i.e. ramps, bridges).
                                    if random.random() < hill_threshold:
                                        lst_strip_sets.append((y, strip_start, strip_end)) 
                            # Reset the strip variables so we can detect another strip.
                            strip_start = -1
                            strip_end = -1
                    else:
                        # Other symbols indicate a break in the current vertical strip.
                        if strip_start != -1:
                            strip_end = x-1
                            # A strip was started in this column, let's review it's length.
                            if (strip_end-strip_start) >= strip_length:
                                # Add this range as a candidate for high_low modification (i.e. ramps, bridges).
                                if random.random() < hill_threshold:
                                    lst_strip_sets.append((y, strip_start, strip_end)) 
                        # Reset the strip variables so we can detect another strip.
                        strip_start = -1
                        strip_end = -1
            # Check to see if a strip was in progress when we reached the edge.
            if strip_start != -1:
                strip_end = x-1
                # A strip was started in this column, let's review it's length.
                if (strip_end-strip_start) >= strip_length:
                    # Add this range as a candidate for high_low modification (i.e. ramps, bridges).
                    if random.random() < hill_threshold:
                        lst_strip_sets.append((y, strip_start, strip_end))
        return lst_strip_sets

    def isTileIDVertical (self, tile_index, id_passed, collection):
        c = 0
        result = False
        entry = collection[tile_index]
        if entry.id_top == id_passed: c += 1
        if entry.id_bottom == id_passed: c += 1
        if c == 2: result = True
        return result
 
    def isTileIDHorizontal (self, tile_index, id_passed, collection):
        c = 0
        result = False
        entry = collection[tile_index]
        if entry.id_left == id_passed: c += 1
        if entry.id_right == id_passed: c += 1
        if c == 2: result = True
        return result
        
    def isTileIDIntersection (self, tile_index, id_passed, collection):
        c = 0
        result = False
        if self.isTileIDVertical(tile_index, id_passed, collection): c += 1
        if self.isTileIDHorizontal(tile_index, id_passed, collection): c += 1
        if c == 2: result = True
        return result
       
    def doesTileContainID (self, tile_index, id_passed, collection):
        result = False
        entry = collection[tile_index]
        if entry.id_top == id_passed: result = True
        if entry.id_bottom == id_passed: result = True
        if entry.id_left == id_passed: result = True
        if entry.id_right == id_passed: result = True
        return result
        
    def returnTileIDSide (self,tile_index, id_passed, collection):
        # Really only works if tile only has a single connection side. (i.e. house connect).
        result = ""
        entry = collection[tile_index]
        if entry.id_top == id_passed: result = "top"
        if entry.id_bottom == id_passed: result = "bottom"
        if entry.id_left == id_passed: result = "left"
        if entry.id_right == id_passed: result = "right"
        return result   

    def returnHouseTileToMatch (self, top, bottom, left, right, house_list, collection, match_required = 4):
        house_connect = "24"
        max_retries = 512
        l = len(house_list)
        if l > 0:
            while max_retries > 0:
                connection_match = 0
                rnd_index = int(random.random() * l)
                index_candidate = house_list[rnd_index]
                connect_side = self.returnTileIDSide (index_candidate, house_connect, collection)
                if connect_side == "top":
                    if top != -1:
                        # We have a valid top tile review connections on.
                        if collection[top].id_bottom == house_connect:
                            # House connects on top and top neighbor connects on bottom, MATCH!
                            return index_candidate
                if connect_side == "bottom":
                    if bottom != -1:
                        # We have a valid bottom tile review connections on.
                        if collection[bottom].id_top == house_connect:
                            # House connects on bottom and bottom neighbor connects on top, MATCH!
                            return index_candidate
                if connect_side == "left":
                    if left != -1:
                        # We have a valid left tile review connections on.
                        if collection[left].id_right == house_connect:
                            # House connects on left and left neighbor connects on right, MATCH!
                            return index_candidate
                if connect_side == "right":
                    if right != -1:
                        # We have a valid right tile review connections on.
                        if collection[right].id_left == house_connect:
                            # House connects on right and right neighbor connects on left, MATCH!
                            return index_candidate
                max_retries -= 1
        return -1 
 
    def returnTileToMatch (self, top, bottom, left, right, grid, collection, match_required = 4):
        max_retries = 512
        l = len(collection)
        if l > 0:
            while max_retries > 0:
                connection_match = 0
                rnd_index = int(random.random() * l)
                entry = collection[rnd_index]
                if entry.enabled:
                    if entry.is_2x_asset == False :
                        try_again = False
                        if self.doesTileContainID (rnd_index,"03",collection):  try_again = True    # Skip road low.
                        if self.doesTileContainID (rnd_index,"04",collection):  try_again = True    # Skip road high.
                        if self.doesTileContainID (rnd_index,"15",collection):  try_again = True    # Skip canal high.
                        if self.doesTileContainID (rnd_index,"16",collection):  try_again = True    # Skip canal low.
                        if self.doesTileContainID (rnd_index,"05",collection):  try_again = True    # Skip water low.
                        if self.doesTileContainID (rnd_index,"06",collection):  try_again = True    # Skip water high.
                        if self.doesTileContainID (rnd_index,"24",collection):                      # 24 is house_connect.
                            # Randomly skip house connections to limit the number of houses.
                            if random.random() > 0.1:
                                try_again = True
                            else:
                                try_again = False
                        if try_again == False:
                            #to_console("Comparing [%s] to [%s]." % (entry.id_top,collection[top].id_top))
                            if top == -1:
                                # This is essentially a wild card so any thing can match.
                                connection_match += 1
                            else:
                                # Does our random selection match our connection type?
                                if entry.id_top == collection[top].id_bottom: connection_match += 1
                            if bottom == -1:
                                # This is essentially a wild card so any thing can match.
                                connection_match += 1
                            else:
                                # Does our random selection match our connection type?
                                if entry.id_bottom == collection[bottom].id_top: connection_match += 1
                            if left == -1:
                                # This is essentially a wild card so any thing can match.
                                connection_match += 1
                            else:
                                # Does our random selection match our connection type?
                                if entry.id_left == collection[left].id_right: connection_match += 1
                            if right == -1:
                                # This is essentially a wild card so any thing can match.
                                connection_match += 1
                            else:
                                # Does our random selection match our connection type?
                                if entry.id_right == collection[right].id_left: connection_match += 1
                            if connection_match == match_required:
                                #to_console ("Found match in %i tries." % (513-max_retries))
                                return rnd_index
                    else:
                        #print ("Skipping tile [%s] is 2x asset." % entry.ob_name)
                        pass
                else:
                    #print ("Skipping tile [%s] is NOT enabled." % entry.ob_name)
                    pass
                max_retries -= 1
        return -1
        
    def isVerticalHighCanal (self, tile_index, collection):
        # Must be a high_canal on top and bottom and no canal connections on left or right.
        result = False
        connection_match = 0
        entry = collection[tile_index]
        if entry.id_top == "15" and  entry.id_bottom == "15": connection_match += 1
        if entry.id_left == "16" or entry.id_left == "15": connection_match -= 1
        if entry.id_right == "16" or entry.id_right == "15": connection_match -= 1
        if connection_match > 0: result = True
        return result
        
    def isHorizontalHighCanal (self, tile_index, collection):
        # Must be a high_canal on left and right and no canal connections on top and bottom.
        result = False
        connection_match = 0
        entry = collection[tile_index]
        if entry.id_left == "15" and  entry.id_right == "15": connection_match += 1
        if entry.id_top == "16" or entry.id_top == "15": connection_match -= 1
        if entry.id_bottom == "16" or entry.id_bottom == "15": connection_match -= 1
        if connection_match > 0: result = True
        return result
       
    def isVerticalHighRiver (self, tile_index, collection):
        # Must be a high_river on top and bottom and no river connections on left or right.
        result = False
        connection_match = 0
        entry = collection[tile_index]
        if entry.id_top == "06" and  entry.id_bottom == "06": connection_match += 1
        if entry.id_left == "05" or entry.id_left == "06": connection_match -= 1
        if entry.id_right == "05" or entry.id_right == "06": connection_match -= 1
        if connection_match > 0: result = True
        return result
        
    def isHorizontalHighRiver (self, tile_index, collection):
        # Must be a high_river on left and right and no river connections on top and bottom.
        result = False
        connection_match = 0
        entry = collection[tile_index]
        if entry.id_left == "06" and  entry.id_right == "06": connection_match += 1
        if entry.id_top == "05" or entry.id_top == "06": connection_match -= 1
        if entry.id_bottom == "05" or entry.id_bottom == "06": connection_match -= 1
        if connection_match > 0: result = True
        return result
        
    def isVerticalHighRoad (self, tile_index, collection):
        # Must be a high_road on top and bottom and no road connections on left or right.
        result = False
        connection_match = 0
        entry = collection[tile_index]
        if entry.id_top == "04" and  entry.id_bottom == "04": connection_match += 1
        if entry.id_left == "03" or entry.id_left == "04": connection_match -= 1
        if entry.id_right == "03" or entry.id_right == "04": connection_match -= 1
        if connection_match > 0: result = True
        return result
        
    def isHorizontalHighRoad (self, tile_index, collection):
        # Must be a high_road on left and right and no road connections on top and bottom.
        result = False
        connection_match = 0
        entry = collection[tile_index]
        if entry.id_left == "04" and  entry.id_right == "04": connection_match += 1
        if entry.id_top == "03" or entry.id_top == "04": connection_match -= 1
        if entry.id_bottom == "03" or entry.id_bottom == "04": connection_match -= 1
        if connection_match > 0: result = True
        return result
        
    def isRoadCliff (self, top, bottom, left, right, tile_index, collection):
        # A road cliff is when a road goes from low (03) to high (04) with no hill/ramp in between.
        result = False
        cliff_match = 0
        entry = collection[tile_index]
        #print ("%s,%s,%s,%s for #%i %s" % (collection[top].id_top,collection[bottom].id_bottom,collection[left].id_left,collection[right].id_right, tile_index, collection[tile_index].ob_name))
        #print(type(entry.id_top),entry.id_top)
        if entry.id_top == "03" and collection[top].id_bottom == "04": cliff_match += 1
        if entry.id_bottom == "03" and collection[bottom].id_top == "04": cliff_match += 1
        if entry.id_left == "03" and collection[left].id_right == "04": cliff_match += 1
        if entry.id_right == "03" and collection[right].id_left == "04": cliff_match += 1
        
        if entry.id_top == "04" and collection[top].id_bottom == "03": cliff_match += 1
        if entry.id_bottom == "04" and collection[bottom].id_top == "03": cliff_match += 1
        if entry.id_left == "04" and collection[left].id_right == "03": cliff_match += 1
        if entry.id_right == "04" and collection[right].id_left == "03": cliff_match += 1
        if cliff_match > 1: result = True
        return result

    def returnMatchForNeighbors (self, top, bottom, left, right, collection):
        top_entry = collection[top]
        bottom_entry = collection[bottom]
        left_entry = collection[left]
        right_entry = collection[right]
        for (i,tile_entry) in enumerate(collection):
            c = 0
            if tile_entry.id_top == top_entry.id_bottom: c += 1
            if tile_entry.id_bottom == bottom_entry.id_top: c += 1
            if tile_entry.id_left == left_entry.id_right: c += 1
            if tile_entry.id_right == right_entry.id_left: c += 1
            if c == 4:
                # 100% match on all sides.
                #print ("matched index #%i to [%s, %s, %s, %s]." % (i,tile_entry.id_top,tile_entry.id_bottom,tile_entry.id_left,tile_entry.id_right))
                return i
        return -1
        
    def patchRoadCliffs (self, solve_path_coordinates, grid, collection):
        for (x,y) in solve_path_coordinates:
            current_tile = collection[grid[x][y]]
            top, bottom, left, right = returnTileNeighborTypes(grid, x, y)
            tile_top = collection[top]
            tile_bottom = collection[bottom]
            tile_left = collection[left]
            tile_right = collection[right]
            # A 04 next to a 03 means a road cliff exists in the grid.                      
            if current_tile.id_top == "04":
                if tile_top.id_bottom == "03":
                    t_id = tile_top.id_top
                    b_id = tile_top.id_bottom
                    l_id = tile_top.id_left
                    r_id = tile_top.id_right
                    if t_id == "24": t_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if b_id == "24": b_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if l_id == "24": l_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if r_id == "24": r_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    b_id = "04" # replace what is currently here with what we are trying to match on.
                    replace_index = self.returnTileMatchFor(t_id,b_id,l_id, r_id, collection)
                    if replace_index != -1:
                        grid[x][y-1] = replace_index
                    else:
                        print ("Unable to find a road cliff TOP match @ (%ix%i)." % (x,y))
                        print(t_id,b_id,l_id,r_id)

            if current_tile.id_bottom == "04":
                if tile_bottom.id_top == "03":
                    t_id = tile_bottom.id_top
                    b_id = tile_bottom.id_bottom
                    l_id = tile_bottom.id_left
                    r_id = tile_bottom.id_right
                    if t_id == "24": t_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if b_id == "24": b_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if l_id == "24": l_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if r_id == "24": r_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    t_id = "04" # replace what is currently here with what we are trying to match on.
                    replace_index = self.returnTileMatchFor(t_id,b_id,l_id, r_id, collection)
                    if replace_index != -1:
                        grid[x][y+1] = replace_index
                    else:
                        print ("Unable to find a road cliff BOTTOM match @ (%ix%i)." % (x,y))

            if current_tile.id_left == "04":
                if tile_left.id_right == "03":
                    t_id = tile_left.id_top
                    b_id = tile_left.id_bottom
                    l_id = tile_left.id_left
                    r_id = tile_left.id_right
                    if t_id == "24": t_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if b_id == "24": b_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if l_id == "24": l_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if r_id == "24": r_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    r_id = "04" # replace what is currently here with what we are trying to match on.
                    replace_index = self.returnTileMatchFor(t_id,b_id,l_id, r_id, collection)
                    if replace_index != -1:
                        grid[x-1][y] = replace_index
                    else:
                        print ("Unable to find a road cliff LEFT match @ (%ix%i)." % (x,y))

            if current_tile.id_right == "04":
                if tile_right.id_left == "03":
                    t_id = tile_right.id_top
                    b_id = tile_right.id_bottom
                    l_id = tile_right.id_left
                    r_id = tile_right.id_right
                    if t_id == "24": t_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if b_id == "24": b_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if l_id == "24": l_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    if r_id == "24": r_id = "01"    # treat a house connect as a low_grass for matching purposes.
                    l_id = "04" # replace what is currently here with what we are trying to match on.
                    replace_index = self.returnTileMatchFor(t_id,b_id,l_id, r_id, collection)
                    if replace_index != -1:
                        grid[x+1][y] = replace_index
                    else:
                        print ("Unable to find a road cliff RIGHT match @ (%ix%i)." % (x,y))

                    
    def removeRoadCliffs (self, grid, collection):
        print("removeRoadCliffs")
        w = len(grid)
        h = len(grid[0])
        for x in range(0,w):
            for y in range(0,h):
                top, bottom, left, right = returnTileNeighborTypes(grid, x, y)
                if self.isRoadCliff (top, bottom, left, right, grid[x][y], collection):
                    # If it is a road cliff, make it a special -3hole, we will patch it later.
                    print("detected a road cliff @%ix%i" % (x,y))
                    grid[x][y] = -3
                    
    def isTileOrphan (self, top, bottom, left, right, tile_index, collection):
        # An orphan is considered a tile with only one matching side.
        result = True
        connection_match = 0
        entry = collection[tile_index]
        if entry.id_top == collection[top].id_bottom: connection_match += 1
        if entry.id_bottom == collection[bottom].id_top: connection_match += 1
        if entry.id_left == collection[left].id_right: connection_match += 1
        if entry.id_right == collection[right].id_left: connection_match += 1
        if connection_match > 1: result = False
        return result
        
    def removeOrphans (self, grid, collection):
        w = len(grid)
        h = len(grid[0])
        for x in range(0,w):
            for y in range(0,h):
                top, bottom, left, right = returnTileNeighborTypes(grid, x, y)
                if self.isTileOrphan (top, bottom, left, right, grid[x][y], collection):
                    # If it is an orphan, make it a hole we will patch it later.
                    grid[x][y] = -1

    def returnFramableList(self,collection):
        # Used by the border option.
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.id_border != '00':
                    result.append(index)
        return result
        
    def returnBorderIndexForMatchingType(self, passed_type, passed_candidates, collection):
        for candidate in passed_candidates:
            entry = collection[candidate]
            if entry.id_border == passed_type:
                # This is the one we are looking for.
                return candidate
        return None
        
    def addBorder (self, grid, collection):
        # See if we even have any border elements in the list.
        w = len(grid)-1
        h = len(grid[0])-1
        lst_border_tiles = self.returnFramableList(collection)
        l = len(lst_border_tiles)
        if l == 6:
            # The exact number we need to build a border around this grid.
            # Border types.
            #("01","Upper Left","ul"),
            #("02","Upper Right","ur"),
            #("03","Lower Left","ll"),
            #("04","Lower Right","lr"),
            #("05","Horizontal","hrz"),
            #("06","Vertical","vert"),
            if w+h > 1:
                ndx = self.returnBorderIndexForMatchingType("05",lst_border_tiles,collection)
                if ndx:
                    for i in range(0,w):
                        grid[i][0] = ndx        #horizontal
                        grid[i][h] = ndx        #horizontal
                ndx = self.returnBorderIndexForMatchingType("06",lst_border_tiles,collection)
                if ndx:
                    for i in range(0,h):
                        grid[0][i] = ndx        #vertical
                        grid[w][i] = ndx        #vertical
                ndx = self.returnBorderIndexForMatchingType("01",lst_border_tiles,collection)
                if ndx: grid[0][0] = ndx        #UL
                ndx = self.returnBorderIndexForMatchingType("02",lst_border_tiles,collection)
                if ndx: grid[w][0] = ndx        #UR
                ndx = self.returnBorderIndexForMatchingType("03",lst_border_tiles,collection)
                if ndx: grid[0][h] = ndx        #LL
                ndx = self.returnBorderIndexForMatchingType("04",lst_border_tiles,collection)
                if ndx: grid[w][h] = ndx        #LR
                
            else:
                # Grid too small to add a border to.
                pass
        elif l == 1:
            # Use only one tile for all border sides.
            if w+h > 1:
                ndx = lst_border_tiles[0]
                for i in range(0,w):
                    grid[i][0] = ndx        #horizontal
                    grid[i][h] = ndx        #horizontal

                for i in range(0,h):
                    grid[0][i] = ndx        #vertical
                    grid[w][i] = ndx        #vertical
                grid[w][h] = ndx
        else:
            if l > 0: print("%i is the wrong number of border tiles defined in list. Shoud be 6 or 1." % l)
            pass
       
    def returnBuildableList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.is_buildable:
                    result.append(index)
        return result
        
    def returnWaterList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.ob_name.find('water') != -1:
                    c = 0
                    if entry.id_top == "05": c +=1
                    if entry.id_bottom == "05": c +=1
                    if entry.id_left == "05": c +=1
                    if entry.id_right == "05": c +=1
                    if c == 4:
                        result.append(index)
        return result
        
    def returnDirtList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.ob_name.find('dirt') != -1:
                    c = 0
                    if entry.id_top == "08": c +=1
                    if entry.id_bottom == "08": c +=1
                    if entry.id_left == "08": c +=1
                    if entry.id_right == "08": c +=1
                    if c == 4:
                        result.append(index)
        return result
        
    def returnSandList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.ob_name.find('sand') != -1:
                    c = 0
                    if entry.id_top == "07": c +=1
                    if entry.id_bottom == "07": c +=1
                    if entry.id_left == "07": c +=1
                    if entry.id_right == "07": c +=1
                    if c == 4:
                        result.append(index)
        return result
       
    def returnGrassList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.ob_name.find('grass') != -1:
                    c = 0
                    if entry.id_top == "01": c +=1
                    if entry.id_bottom == "01": c +=1
                    if entry.id_left == "01": c +=1
                    if entry.id_right == "01": c +=1
                    if c == 4:
                        result.append(index)
        return result

    def returnHillList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                is_candidate = False
                if (entry.ob_name.find('hill') != -1): is_candidate = True
                # Filter out hills and ramps.
                if (entry.ob_name.find('bridge') != -1): is_candidate = False
                if (entry.ob_name.find('house') != -1): is_candidate = False
                if (entry.ob_name.find('river') != -1): is_candidate = False
                if (entry.ob_name.find('canal') != -1): is_candidate = False
                if (entry.ob_name.find('road') != -1): is_candidate = False
                if (entry.ob_name.find('dirt') != -1): is_candidate = False
                if (entry.ob_name.find('sand') != -1): is_candidate = False
                if is_candidate:
                    result.append(index)
        return result

    def returnDuneList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                is_candidate = False
                if (entry.ob_name.find('dune') != -1): is_candidate = True
                if is_candidate:
                    result.append(index)
        return result
        
    def returnCanalList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                is_candidate = False
                if (entry.ob_name.find('canal') != -1): is_candidate = True
                # Filter out hills and ramps.
                if (entry.ob_name.find('hill') != -1): is_candidate = False
                if (entry.ob_name.find('ramp') != -1): is_candidate = False
                if is_candidate:
                    result.append(index)
        return result
        
    def returnRiverList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                is_candidate = False
                if (entry.ob_name.find('water') != -1): is_candidate = True
                if (entry.ob_name.find('river') != -1): is_candidate = True
                # Filter out roads, bridges and hills.
                if (entry.ob_name.find('road') != -1): is_candidate = False
                if (entry.ob_name.find('bridge') != -1): is_candidate = False
                if (entry.ob_name.find('hill') != -1): is_candidate = False
                if is_candidate:
                    result.append(index)
        return result
 
    def returnTreeList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.is_plant:
                    if entry.ob_name.find('tree') != -1:
                            result.append(index)
        return result
        
    def returnBushList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.is_plant:
                    if entry.ob_name.find('bush') != -1:
                            result.append(index)
        return result
        
    def returnFlowerList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.is_plant:
                    if entry.ob_name.find('flower') != -1:
                            result.append(index)
        return result
        
    def returnStreetAssetList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.is_street_asset:
                    result.append(index)
        return result
        
    def returnAdvertisingAssetList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.is_advertising_asset:
                    result.append(index)
        return result
        
    def returnRoadList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.ob_name.find('road') != -1:
                    if entry.ob_name.find('ramp') != -1:
                        # Skip ramps.
                        pass
                    else:
                        if entry.ob_name.find('bridge') != -1:
                            # Skip bridge tiles.
                            pass
                        else:
                            result.append(index)
        return result

    def returnHighRoadList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.ob_name.find('road') != -1:
                    if (entry.ob_name.find('high') != -1) or (entry.ob_name.find('bridge') != -1):
                        if (entry.ob_name.find('mid') != -1):
                            # Skip experimental mid slope connector types.
                            pass
                        else:
                            # Add road high tiles.
                            result.append(index)
        return result

    def convertHouseConnectsAlongSolvePath(self, passed_map, grid, collection):
        # There is no match for the high road house connect so convert those types into standard road tiles which can be matched.
        w = len(grid)
        h = len(grid[0])
        for x in range(0,w):
            for y in range(0,h):
                if passed_map[x][y] == ".":
                    entry = collection[grid[x][y]]
                    if (entry.ob_name.find('road-straight-house-1-low_R') != -1) or (entry.ob_name.find('road-straight-house-1-low_L') != -1) or (entry.ob_name.find('road-straight-house-2-low_TB') != -1):
                        grid[x][y] = 210    # HARD CODED non_house_connect_road
                    if (entry.ob_name.find('road-straight-house-1-low_B') != -1) or (entry.ob_name.find('road-straight-house-1-low_T') != -1) or (entry.ob_name.find('road-straight-house-2-low_LR') != -1):
                        grid[x][y] = 209    # HARD CODED non_house_connect_road
        
    def returnHouseList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.ob_name.find('house') != -1:
                    if entry.ob_name.find('ramp') != -1:
                        # Skip ramps.
                        pass
                    else:
                        if entry.ob_name.find('bridge') != -1:
                            # Skip bridge tiles.
                            pass
                        else:
                            if entry.ob_name.find('road') != -1:
                                # Skip road tiles.
                                pass
                            else:
                                result.append(index)
        return result
        
    def returnPlantList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.ob_name.find('plant') != -1:
                    result.append(index)
        return result
                    
    def returnFillerList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.is_filler:
                    result.append(index)
        return result
        
    def return2xAssetList(self,collection):
        result = []
        for (index, entry) in enumerate(collection):
            if entry.enabled:
                if entry.is_2x_asset:
                    result.append(index)
        return result
       
    def patchWithForgivingConnections (self, grid, collection):
        # At this point we have some holes that just will not connect on all four side.
        # So let's try matching on 3, 2, then 1 sides to fill all holes.
        w = len(grid)
        h = len(grid[0])
        for x in range(0,w):
            for y in range(0,h):
                top, bottom, left, right = returnTileNeighborTypes(grid, x, y)
                if grid[x][y] == -1:
                    tile_id = self.returnTileToMatch (top, bottom, left, right, grid, collection, 3)
                    if tile_id == -1:
                        tile_id = self.returnTileToMatch (top, bottom, left, right, grid, collection,2)
                        if tile_id == -1:
                            tile_id = self.returnTileToMatch (top, bottom, left, right, grid, collection, 1)
                            if tile_id == -1:
                                # Still no match...make it filler.
                                lst = self.returnFillerList(collection)
                                if len(lst):
                                    n = random.randint(0,(len(lst)-1))
                                    to_console(collection[n].ob_name)
                                    tile_id = lst[n]
                                else:
                                    to_console ("No filler tile specified, leaving hole in grid.")
                    grid[x][y] = tile_id
                    
    def populateGridHoles (self, grid, collection):
        w = len(grid)
        h = len(grid[0])
        for x in range(0,w):
            for y in range(0,h):
                if grid[x][y] == -1:
                    # This is a hole in the grid.
                    tile_id = -1
                    top, bottom, left, right = returnTileNeighborTypes(grid, x, y)
                    #to_console ("Neighbors @%ix%i -> T:%i B:%i L:%i R:%i" % (x,y,top,bottom,left,right))
                    tile_id = self.returnTileToMatch (top, bottom, left, right, grid, collection)
                    if tile_id != -1:
                        #to_console ("Match candidate @%ix%i -> T:%s B:%s L:%s R:%s" % (x,y,collection[tile_id].id_top,collection[tile_id].id_bottom,collection[tile_id].id_left,collection[tile_id].id_right))
                        grid[x][y] = tile_id
                    else:
                        pass
                        #to_console ("Leaving hole at [%ix%i], no suitable tile match found." % (x,y))
                        
    def populateGridWithTile (self, probability, tile_id, grid, collection):
        w = len(grid)
        h = len(grid[0])
        for x in range(0,w):
            for y in range(0,h):
                if grid[x][y] == -1:
                    # This is a blank space that can be populated.
                    probability_check = random.random()
                    if probability > probability_check:
                        # Add a tile.
                        grid[x][y] = tile_id
        
    def returnGridFromMap(self, passed_map, passed_candidates, match_types, collection):
        w = len(passed_map)
        h = len(passed_map[0])
        grid = self.returnBlankGrid(w,h)
        for x in range(0,w):
            for y in range(0,h):
                if passed_map[x][y] == ".":
                    # An empty space in the passed map.
                    top, bottom, left, right = returnMapNeighborTypes(passed_map, x, y)
                    connection_type = returnSelfDeterminedByNeighbors(top,bottom,left,right)
                    for ndx in passed_candidates:
                        tile_id = -1    # i.e. no match.
                        entry = collection[ndx]
                        # Use brute force matching....ugly but just might work.
                        skip_crossroads = False
                        if x == 0: skip_crossroads = True
                        if x == (w-1): skip_crossroads = True
                        if y == 0: skip_crossroads = True
                        if y == (h-1): skip_crossroads = True
                        if skip_crossroads == False:
                            # Crossroads intersection.
                            if connection_type == 'XX' and (entry.id_top in match_types) and (entry.id_bottom in match_types) and (entry.id_left in match_types) and (entry.id_right in match_types): 
                                tile_id = ndx
                                break
                        # T intersections.  
                        if connection_type == 'TT' and (entry.id_top not in match_types) and (entry.id_bottom in match_types) and (entry.id_left in match_types) and (entry.id_right in match_types): 
                            tile_id = ndx
                            break
                        if connection_type == 'TB' and (entry.id_top in match_types) and (entry.id_bottom not in match_types) and (entry.id_left in match_types) and (entry.id_right in match_types): 
                            tile_id = ndx
                            break
                        if connection_type == 'TL' and (entry.id_top in match_types) and (entry.id_bottom in match_types) and (entry.id_left not in match_types) and (entry.id_right in match_types): 
                            tile_id = ndx
                            break
                        if connection_type == 'TR' and (entry.id_top in match_types) and (entry.id_bottom in match_types) and (entry.id_left in match_types) and (entry.id_right not in match_types) : 
                            tile_id = ndx
                            break
                        # Corner types.
                        if connection_type == 'BR' and (entry.id_top not in match_types) and (entry.id_bottom in match_types) and (entry.id_left not in match_types) and (entry.id_right in match_types): 
                            tile_id = ndx
                            break
                        if connection_type == 'BL' and (entry.id_top not in match_types) and (entry.id_bottom in match_types) and (entry.id_left in match_types) and (entry.id_right not in match_types) : 
                            tile_id = ndx
                            break
                        if connection_type == 'RT' and (entry.id_top in match_types) and (entry.id_bottom not in match_types) and (entry.id_left not in match_types) and (entry.id_right in match_types): 
                            tile_id = ndx
                            break
                        if connection_type == 'LT' and (entry.id_top in match_types) and (entry.id_bottom not in match_types) and (entry.id_left in match_types) and (entry.id_right not in match_types) : 
                            tile_id = ndx
                            break
                        # Vertical, horizontal.
                        if connection_type == '==' and (entry.id_top not in match_types) and (entry.id_bottom not in match_types) and (entry.id_left in match_types) and (entry.id_right in match_types): 
                            tile_id = ndx
                            break
                        if connection_type == '||' and (entry.id_top in match_types) and (entry.id_bottom in match_types) and (entry.id_left not in match_types) and (entry.id_right not in match_types) : 
                            tile_id = ndx
                            break
                        # Single types.
                        if connection_type == 'ST' and (entry.id_top in match_types) and (entry.id_bottom not in match_types) and (entry.id_left not in match_types) and (entry.id_right not in match_types) : 
                            tile_id = ndx
                            break
                        if connection_type == 'SB' and (entry.id_top not in match_types) and (entry.id_bottom in match_types) and (entry.id_left not in match_types) and (entry.id_right not in match_types) : 
                            tile_id = ndx
                            break
                        if connection_type == 'SL' and (entry.id_top not in match_types) and (entry.id_bottom not in match_types) and (entry.id_left in match_types) and (entry.id_right not in match_types) : 
                            tile_id = ndx
                            break
                        if connection_type == 'SR' and (entry.id_top not in match_types) and (entry.id_bottom not in match_types) and (entry.id_left not in match_types) and (entry.id_right in match_types): 
                            tile_id = ndx
                            break
                    grid[x][y] = tile_id
                else:
                    # Choose randomly from items in the filler list.
                    #grid[x][y] = lst_filler[random.randint(0,len(lst_filler)-1)]
                    pass
        return grid

    def returnRandomGridOverArea (self, width, height, candidate_list, area_map, rnd_seed):
        # Return a grid with a random candidate only where the area_map has a space.
        random.seed(rnd_seed)
        grid = self.returnBlankGrid(width,height)
        l = len(candidate_list)
        if l > 0:
            for x in range(0,width):
                for y in range(0,height):
                    if area_map[x][y] == '.':
                        # This space is empty and can be randomly populated.
                        ndx = random.randint(0,(l-1))
                        grid[x][y] = candidate_list[ndx]            # The index into the tile list entry that contains the TBLR (top, bottom, left, right) information as well as the name of the object.
        return grid

    def fillDeadEnds (self, passed_map):
        w = len(passed_map)
        h = len(passed_map[0])
        grid = self.returnBlankGrid(w,h)
        # Copy passed map to our grid.
        # We are going to modifyit and return it.
        for x in range(0,w):
                for y in range(0,h):
                    grid[x][y] = passed_map[x][y]
        break_loop = False
        while break_loop == False:
            #Keep searching map for dead ends till none are found.
            dead_end_count = 0
            print("\n")
            for x in range(0,w):
                for y in range(0,h):
                    if grid[x][y] == "#":
                        # Already a solid wall, nothing to review.
                        pass
                    else:
                        c = 0
                        top,bottom,left,right = returnMapNeighborTypes(grid,x,y)
                        print ("%s%s%s%s (%ix%i)" % (top,bottom,left,right,x,y))
                        if top == ".": c+=1
                        if bottom == ".": c+=1
                        if left == ".": c+=1
                        if right == ".": c+=1
                        if c < 2:
                            # This is a dead end and should be filled.
                            grid[x][y] = "#"
                            dead_end_count +=1
            if dead_end_count == 0:
                break_loop = True
            #break_loop = True
        return grid

    def recursiveSolve(self, maze, was_here, correct_path, x, y, endX, endY, width, height, solution_coordinates):
        if x == endX and y == endY:
            # If you reached the end
            return True 
        
        if maze[x][y] == "#" or was_here[x][y]: 
            # If you are on a wall or already were here
            return False
     
        was_here[x][y] = True 
        if x != 0:              # Checks if not on left edge
            if self.recursiveSolve(maze, was_here, correct_path, x-1, y, endX, endY, width, height, solution_coordinates):       #Recalls method one to the left
                correct_path[x][y] = "."    #Sets that path value to true;
                solution_coordinates.append((x,y))
                return True

        if x != width - 1:      #Checks if not on right edge
            if self.recursiveSolve(maze, was_here, correct_path, x+1, y, endX, endY, width, height, solution_coordinates):       #Recalls method one to the right
                correct_path[x][y] = "."
                solution_coordinates.append((x,y))
                return True

        if y != 0:              #Checks if not on top edge
            if self.recursiveSolve(maze, was_here, correct_path, x, y-1, endX, endY, width, height, solution_coordinates):       #Recalls method one up
                correct_path[x][y] = "."
                solution_coordinates.append((x,y))
                return True

        if y != height- 1:      #Checks if not on bottom edge
            if self.recursiveSolve(maze, was_here, correct_path, x, y+1, endX, endY, width, height, solution_coordinates):       #Recalls method one down
                correct_path[x][y] = "."
                solution_coordinates.append((x,y))
                return True
        return False 
        
    def solveMaze (self, passed_map):
        w = len(passed_map)
        h = len(passed_map[0])
        solution_coordinates = []                       # An array of x,y coordinates in the order that the maze was solved.
        was_here = self.returnBlankGrid(w,h, False)     # Entire map starts off False.  
        correct_path = self.returnBlankGrid(w,h, "#")   # Solve map starts off all walls.
        
        # Determine start and end points.
        startX = int(random.random() * w)
        startY = int(random.random() * h)
        max_retries = 512
        while passed_map[startX][startY] == "#":
            startX = int(random.random() * w)
            startY = int(random.random() * h)
            max_retries -= 1
            if max_retries < 0: break
        
        # Try to make the end point fairly far away from the start point.
        endX = int(random.random() * w)
        endY = int(random.random() * h)
        distX = abs(endX-startX)
        distY = abs(endY-startY)
        end_dist = distX
        end_compare = int(w/2)
        if distY > distX: 
            end_dist = distY
            end_compare = int(h/2)
        max_retries = 512
        while passed_map[endX][endY] == "#" or (end_dist < end_compare):
            endX = int(random.random() * w)
            endY = int(random.random() * h)
            end_dist = abs(endX-startX)
            max_retries -= 1
            if max_retries < 0: break
            
        #print("(%ix%i) (%ix%i)" % (startX, startY, endX, endY))
        r = self.recursiveSolve(passed_map, was_here, correct_path, startX, startY, endX, endY, w, h, solution_coordinates)
        return correct_path, solution_coordinates

    def returnMatchToTileFromCandidates(self, top_entry, bottom_entry, left_entry, right_entry, lst_candidates, collection, match_count):
        tile_id = -1
        max_retries = 512
        l = len(lst_candidates)
        if l > 0:
            while max_retries > 0:
                #print ("attempting to match on try #%i." % max_retries)
                rnd_index = int(random.random() * l)
                candidate_entry = collection[lst_candidates[rnd_index]]
                c = 0
                if candidate_entry.id_top == top_entry.id_bottom: c +=1
                if candidate_entry.id_bottom == bottom_entry.id_top: c +=1
                if candidate_entry.id_left == left_entry.id_right: c +=1
                if candidate_entry.id_right == right_entry.id_left: c +=1
                if  c >= match_count: 
                    #print ("found match in #%i tries." % max_retries)
                    return rnd_index
                max_retries -= 1
        return tile_id

    def returnMatchToTileFromCollection(self, top_entry, bottom_entry, left_entry, right_entry, collection, match_count):
        tile_id = -1
        max_retries = 512
        l = len(collection)
        if l > 0:
            while max_retries > 0:
                rnd_index = int(random.random() * l)
                candidate_entry = collection[rnd_index]
                c = 0
                if candidate_entry.id_top == top_entry.id_bottom: c +=1
                if candidate_entry.id_bottom == bottom_entry.id_top: c +=1
                if candidate_entry.id_left == left_entry.id_right: c +=1
                if candidate_entry.id_right == right_entry.id_left: c +=1
                if  c >= match_count: 
                    return rnd_index
                max_retries -= 1
        return tile_id

    def returnTileMatchFor(self, top_id, bottom_id, left_id, right_id, collection):
        result = -1
        match_candidates = []
        for (i,entry) in enumerate(collection):  
            c = 0
            if entry.id_top == top_id: c+= 1
            if entry.id_bottom == bottom_id: c+= 1
            if entry.id_left == left_id: c+= 1
            if entry.id_right == right_id: c+= 1
            if c == 4:
                match_candidates.append(i)
        l = len (match_candidates)
        if l > 0:
            rnd_index = int(random.random() * l)
            result = match_candidates[rnd_index]
        return result
               
    def addRND2xAssets(self, tile_id, grid, collection):
        w = len(grid)
        h = len(grid[0])
        x = int(random.random() * len(grid))
        y = int(random.random() * len(grid[0]))
        if x > 2 and x < (w-2):
            # Larger assets can not be too near the edge.
            if y > 2 and y < (h-2):
                    grid[x][y] = tile_id
                    grid[x+1][y] = -2
                    grid[x+1][y+1] = -2
                    grid[x][y+1] = -2
             
    def add2xAssets(self, tile_id, grid, collection):
        w = len(grid)
        h = len(grid[0])
        for x in range(0,w):
            if x > 2 and x < (w-2):
                for y in range(0,h):
                    # Larger assets can not be too near the edge.
                    if y > 2 and y < (h-2):
                        lst = self.returnBuildableList(collection)
                        if len(lst):
                            if grid[x][y] in lst:           
                                grid[x][y] = tile_id
                                grid[x+1][y] = -2
                                grid[x+1][y+1] = -2
                                grid[x][y+1] = -2
                                return True
                        else:
                            to_console ("No buildable tile specified..?")
        return False
   
    def invoke(self, context, event):
        ob = context.object
        if ob != None:
            collection = ob.TileProc_List
            l = len(collection)
            if l > 0:
                # Properties from the interface.
                random.seed(int(ob["rnd_seed"]))
                w = int(ob["width"])
                h = int(ob["height"])
                fill_amount = bool(ob["filler_amount"])
                
                # These need to be appended in the correct order. 
                # The order is specified in the enum of cls_TileProcLayer in properties.py.
                match_types = []
                match_types.append(['08'])  # dirt_low.
                match_types.append(['07'])  # sand_low.
                match_types.append(['05'])  # water_low.
                match_types.append(['01'])  # grass_low.
                match_types.append(['02'])  # grass_high.
                match_types.append(['07'])  # sand.
                match_types.append(['21'])  # water_junction.
                match_types.append(['16'])  # canal_low.
                match_types.append(['03'])  # road_low.

                lst_candidates = []
                lst_candidates.append(self.returnDirtList(collection))
                lst_candidates.append(self.returnSandList(collection))
                lst_candidates.append(self.returnWaterList(collection))
                lst_candidates.append(self.returnGrassList(collection))
                lst_candidates.append(self.returnHillList(collection))
                lst_candidates.append(self.returnDuneList(collection))
                lst_candidates.append(self.returnRiverList(collection))
                lst_candidates.append(self.returnCanalList(collection))
                lst_candidates.append(self.returnRoadList(collection))
                lst_candidates.append(self.returnHouseList(collection))
                lst_candidates.append(self.returnTreeList(collection))
                lst_candidates.append(self.returnBushList(collection))
                lst_candidates.append(self.returnFlowerList(collection))
                #lst_candidates.append(self.returnVehicleList(collection))
                
                lst_high_roads = self.returnHighRoadList(collection)
                
                lst_candidates_vertical = []
                lst_candidates_vertical.append([])
                lst_candidates_vertical.append([])
                lst_candidates_vertical.append([])
                lst_candidates_vertical.append([])
                lst_candidates_vertical.append([])
                lst_candidates_vertical.append([])
                lst_candidates_vertical.append(self.returnHighRiversVerticalCandidates(collection))
                lst_candidates_vertical.append(self.returnHighCanalsVerticalCandidates(collection))
                lst_candidates_vertical.append(self.returnHighRoadsVerticalCandidates(collection))
                lst_candidates_vertical.append([])
                lst_candidates_vertical.append([])
                lst_candidates_vertical.append([])

                lst_candidates_horizontal = []
                lst_candidates_horizontal.append([])
                lst_candidates_horizontal.append([])
                lst_candidates_horizontal.append([])
                lst_candidates_horizontal.append([])
                lst_candidates_horizontal.append([])
                lst_candidates_horizontal.append([])
                lst_candidates_horizontal.append(self.returnHighRiversHorizontalCandidates(collection))
                lst_candidates_horizontal.append(self.returnHighCanalsHorizontalCandidates(collection))
                lst_candidates_horizontal.append(self.returnHighRoadsHorizontalCandidates(collection))
                lst_candidates_horizontal.append([])
                lst_candidates_horizontal.append([])
                lst_candidates_horizontal.append([])
                
                # Vertical and Horizontal modifications are only for maze area types.
                # These are HARD CODED numbers into the collection list.
                # They should be derived from connection types instead.
                lst_vertical_connectors = []
                lst_vertical_connectors.append([0,0])   # dirt_low.
                lst_vertical_connectors.append([0,0])   # sand_low.
                lst_vertical_connectors.append([0,0])   # water_low.
                lst_vertical_connectors.append([0,0])   # grass_low.
                lst_vertical_connectors.append([0,0])   # grass_high.
                lst_vertical_connectors.append([0,0])   # sand.
                lst_vertical_connectors.append([141,138])   #river-hill_BT, river-hill_TB
                lst_vertical_connectors.append([21,18])     #canal-hill_BT, canal-hill_TB
                lst_vertical_connectors.append([200,197])   #road-hill_BT, road-hill_TB
                
                lst_horizontal_connectors = []
                lst_horizontal_connectors.append([0,0]) # dirt_low.
                lst_horizontal_connectors.append([0,0]) # sand_low.
                lst_horizontal_connectors.append([0,0]) # water_low.
                lst_horizontal_connectors.append([0,0]) # grass_low.
                lst_horizontal_connectors.append([0,0]) # grass_high.
                lst_horizontal_connectors.append([0,0]) # sand.
                lst_horizontal_connectors.append([139,140]) #river-hill_LR, river-hill_RL
                lst_horizontal_connectors.append([19,20])   #canal-hill_LR, canal-hill_RL
                lst_horizontal_connectors.append([198,199]) #road-hill_LR, road-hill_RL
                
                area_types = ['Dirt','Sand','Grass', 'Water','Hills','Dunes']
                maze_types = ['Rivers','Canals','Roads']
                building_types = ['Houses']
                plant_types = ['Trees', 'Bushes', 'Flowers', 'Street Assets','Advertising Assets']
                vehicle_types = ['Vehicles', 'Birds']
                atmosphere_types = ['Clouds', 'UFOs']
                
                # Construct all the masking maps first.                
                maps = []
                for i,layer in enumerate(ob.TileProcLayers_List):
                    # Build a random area map for this layer.
                    if layer.area_invert:
                        area_map = self.invertMap(self.returnRndAreaMap(w,h, layer.area, layer.area_center_x, layer.area_center_y, (int(ob["rnd_seed"])+layer.rnd_seed)))
                    else:
                        area_map = self.returnRndAreaMap(w,h, layer.area, layer.area_center_x, layer.area_center_y, (int(ob["rnd_seed"])+layer.rnd_seed))

                    maps.append(area_map)
                    if layer.name == 'Clouds':
                        self.printGrid(area_map,"Pre Area [%s]." % layer.name) 
                    
                # Construct a grid for each layer type using previously generated masking maps.
                grids = []
                grid = self.returnBlankGrid(w,h)
                for i,layer in enumerate(ob.TileProcLayers_List):
                    random.seed(int(ob["rnd_seed"])+layer.rnd_seed)
                    n = int(layer.id_map_mask)
                    temp_grid = self.returnBlankGrid(w,h)
                    
                    # Fetch the area map for this layer or the optional one if not None.
                    if layer.id_map_mask != '00':
                        # Use a map to mask this grid. NOTE: index is off by one, zero is self.
                        ndx = n-1
                        #self.printGrid(area_map,"Area Map [%s]." % ob.TileProcLayers_List[n-1].name)
                    else:
                        # '00', None means use self.
                        ndx = i
                    
                    # The area_map for this layer is available even if the layer is not enabled.    
                    if layer.area_invert:
                        area_map = self.invertMap(maps[ndx])
                    else:
                        area_map = maps[ndx]
                    #self.printGrid(area_map,"Area Map From [%s]." % ob.TileProcLayers_List[ndx].name)
                    
                    if layer.enabled:        
                        # Build a grid over an area for this layer based upon the layer name.
                        if layer.name in area_types:
                            temp_grid = self.returnRandomGridOverArea (w, h, lst_candidates[i], area_map, (int(ob["rnd_seed"])+layer.rnd_seed))

                        if layer.name in maze_types:
                            if layer.area_invert:
                                maze_map = self.returnMazeMap(w,h,layer.branch_rate, self.invertMap(area_map),(int(ob["rnd_seed"])+layer.rnd_seed))                     # Mazes flow around the area map.
                            else:
                                maze_map = self.returnMazeMap(w,h,layer.branch_rate, area_map,(int(ob["rnd_seed"])+layer.rnd_seed))                                     # Mazes flow around the area map.
                            maps[i] = maze_map
                            temp_grid = self.returnGridFromMap(maze_map, lst_candidates[i], match_types[i], collection)
                            #self.printGrid(maze_map,"Maze Map [%s] %i." % (layer.name, i))
                            
                            if layer.name == "Roads":
                                #Solve the maze to produce a solution that will be used for high roads.
                                solution_map, solution_coordinates = self.solveMaze(maze_map)
                                #self.printGrid(solution_map,"Solution Map [%s]." % layer.name)
                                # Convert house connects into normal road along solution_map path.
                                self.convertHouseConnectsAlongSolvePath(solution_map, temp_grid, collection)
                                
                                # Apply high roads to the solution map.
                                solve_path_coordinates = []
                                for x in range(0,w):
                                    for y in range(0,h):
                                        if solution_map[x][y] == ".":
                                            solve_path_coordinates.append((x,y))
                                            # Replace this low road section with a high road alternate.
                                            current_tile = collection[temp_grid[x][y]]
                                            current_orientation = self.returnTileOrientationViaType("03",temp_grid[x][y],collection)                    # Determine my current orientation. 03-road_low
                                            if current_orientation != None:
                                                #print("current_tile %s @(%ix%i) has orientation %s" % (current_tile.name,x,y,current_orientation))
                                                max_retries = 512
                                                l = len(lst_high_roads)
                                                if l > 0:
                                                    while max_retries > 0:
                                                        rnd_index = int(random.random() * l)
                                                        high_road_index = lst_high_roads[rnd_index]
                                                        high_road_orientation = self.returnTileOrientationViaType("04",high_road_index,collection)      # Determine my current orientation. 04=road_high
                                                        if high_road_orientation != None:
                                                            if current_orientation == high_road_orientation:
                                                                # This is my replacement.
                                                                #print ("Found high road orientation match. %i" % temp_grid[x][y])
                                                                temp_grid[x][y] = high_road_index
                                                                break
                                                        else:
                                                            pass
                                                            #print ("Unable to determine high road orientation. %i" % temp_grid[x][y])
                                                        max_retries -= 1
                                            else:
                                                print ("Unable to determine low road orientation. %i" % temp_grid[x][y])
                                
                                self.patchRoadCliffs(solve_path_coordinates, temp_grid, collection)
  
                            # Add hills if possible.
                            self.modifyVerticalStrips(layer.threshold, maze_map, temp_grid, lst_candidates_vertical[i], lst_vertical_connectors[i], collection)         # Add some high roads and bridges if possible.
                            self.modifyHorizontalStrips(layer.threshold, maze_map, temp_grid, lst_candidates_horizontal[i], lst_horizontal_connectors[i], collection)   # Add some high roads and bridges if possible.

                        if layer.name in building_types:
                            # Populate houses next to road house connects.
                            road_grid = grids[8]    # HARD CODED reference to slot #8, bad coding.
                            #self.printGrid(road_grid,"#%i Road grid [%s]." % (i,layer.name))
                            w = len(road_grid)
                            h = len(road_grid[0])
                            for x in range(0,w):
                                for y in range(0,h):
                                    if road_grid[x][y] == -1:
                                        if area_map[x][y] == '.':
                                            # Ok to build here.
                                            top, bottom, left, right = returnTileNeighborTypes(road_grid, x, y)
                                            if (top+bottom+left+right) == -4:
                                                # No tiles around to connect to, no need to review further.
                                                pass
                                            else:
                                                ndx = self.returnHouseTileToMatch (top, bottom, left, right, lst_candidates[i], collection, 1)
                                                if ndx != -1:
                                                    temp_grid[x][y] = ndx
                                                    
                        if layer.name in vehicle_types:
                            pass
                    # Populate the master grid with the grid candidates.
                    # Layers stacked one on top of the other replacing what was below.
                    grids.append(temp_grid)
                    for grid_candidate in grids:
                        for x in range(0,w):
                            for y in range(0,h):
                                if temp_grid[x][y] != -1:
                                    grid[x][y] = grid_candidate[x][y]
                                    
                    if layer.name in plant_types:
                        # Distribute plants based upon area_map.
                        house_grid = grids[9]   # HARD CODED reference to slot #9, bad coding.
                        #self.printGrid(house_grid,"#%i House grid [%s]." % (i,layer.name))
                        #self.printGrid(area_map,"#%i ARea Map [%s]." % (i,layer.name))
                        for x in range(0,w):
                            for y in range(0,h):
                                e = layer.enabled
                                if house_grid[x][y] != -1:
                                    # Don't place plants where a house is located.
                                    e = False
                                if e:
                                    self.itemsToTile (context.scene, ob, layer, area_map[x][y], x, y, grid, collection,(int(ob["rnd_seed"])+layer.rnd_seed)) 
                                else:
                                    # Not enabled try to unlink prior count based candidates.
                                    for c in range(layer.item_count+10):
                                        n = "%s_%s" % (ITEM_NAME, layer.name)
                                        item_name = returnItemName(n, x, y, c) # Name of item to be created.
                                        temp_ob_item = bpy.data.objects.get(item_name)
                                        if temp_ob_item != None:
                                            try:
                                                context.scene.objects.unlink(temp_ob_item)      # Unlink from scene if in scene.
                                            except:
                                                pass				# Already unlinked?

                    if layer.name in atmosphere_types:
                        if layer.name == 'Clouds' or layer.name == 'UFOs':
                            random.seed(int(ob["rnd_seed"])+layer.rnd_seed)
                            # Make a list of cloud objects by reviewing the 'clouds' group. HARD CODED reference to a group.
                            if layer.name == 'Clouds':
                                grp = bpy.data.groups.get('clouds')
                            elif layer.name == 'UFOs':
                                grp = bpy.data.groups.get('ufos')
                            candidates = []
                            for grp_ob in grp.objects:
                                candidates.append(grp_ob.name)
                            lc = len(candidates)
                            if layer.enabled:
                                self.printGrid(area_map,"Post Area [%s]." % layer.name)
                                for x in range(0,w):
                                    for y in range(0,h):
                                        if area_map[x][y] == '.':
                                            # Create some clouds at this location.
                                            for c in range(layer.cloud_count):
                                                n = random.randint(0,(lc-1))
                                                cloud_source_name = candidates[n]
                                                ob_cloud_source = bpy.data.objects.get(cloud_source_name)
                                                if ob_cloud_source:
                                                    # Clone the cloud source.    
                                                    new_cloud_name = returnItemName(("%s" % cloud_source_name),x, y, c)
                                                    ob_cloud_new = bpy.data.objects.get(new_cloud_name)
                                                    if ob_cloud_new == None:
                                                        # I don't exist yet, time to create.
                                                        ob_cloud_new = ob_cloud_source.copy()
                                                        ob_cloud_new.name = new_cloud_name
                                                    else:
                                                        ob_cloud_new.data = ob_cloud_source.data
                                                    try:
                                                        context.scene.objects.link(ob_cloud_new)
                                                    except:
                                                        pass

                                                    # Generate a random location within this tile's area which is 8x8 Blender units in size by default.
                                                     #Fetch information from the AddOn preferences.
                                                    unit_size = bpy.context.user_preferences.addons['tile_proc'].preferences.unit_size
                                                    x_center = unit_size * x - (unit_size * (w/2))
                                                    y_center = -unit_size * y - (-unit_size * (h/2))
                                                    area_size = unit_size * layer.cloud_coverage
                                                    x_offset = (random.random() * area_size)
                                                    y_offset = (random.random() * area_size)
                                                    ob_cloud_new.location.x = x_center + (x_offset-(area_size/2))
                                                    ob_cloud_new.location.y = y_center + (y_offset-(area_size/2))
                                                    rnd_z = int(random.random() * (unit_size/2))
                                                    ob_cloud_new.location.z = layer.cloud_height + rnd_z
                                                    rnd_size = (random.random() * 2) * layer.cloud_rnd_size
                                                    ob_cloud_new.scale = (1.0+rnd_size,1.0+rnd_size,1.0+rnd_size)
                                                    if layer.cloud_axis_type == '01':
                                                        ob_cloud_new.rotation_euler.z = 90
                                                    else:
                                                        ob_cloud_new.rotation_euler.z = 0
                                                    try:
                                                        ob_cloud_new.parent = ob
                                                    except:
                                                        pass		# Already parented?
                                        else:
                                            # No clouds should appear in this area.
                                            # We need to make an attempt to remove any previous clouds by scanning for names of known objects.
                                            for c in range(layer.cloud_count+10):
                                                n = random.randint(0,(lc-1))
                                                cloud_source_name = candidates[n]
                                                # Try to unlink prior count based candidates.
                                                new_name = returnItemName(("%s" % cloud_source_name),x, y, c)
                                                try:
                                                    context.scene.objects.unlink(bpy.data.objects.get(new_name))      # Unlink from scene if in scene.
                                                except:
                                                    pass	# Already unlinked?
                            else:
                                # No clouds should appear in this area.
                                for x in range(0,w):
                                    for y in range(0,h):
                                        # We need to make an attempt to remove any previous clouds by scanning for names of known objects.
                                        for c in range(layer.cloud_count+10):
                                            n = random.randint(0,(lc-1))
                                            cloud_source_name = candidates[n]
                                            # Try to unlink prior count based candidates.
                                            new_name = returnItemName(("%s" % cloud_source_name),x, y, c)
                                            try:
                                                context.scene.objects.unlink(bpy.data.objects.get(new_name))      # Unlink from scene if in scene.
                                            except:
                                                pass	# Already unlinked? 
           
                    if layer.name in vehicle_types:		# Vehicle types should be renamed to rig type...I guess?
                        random.seed(int(ob["rnd_seed"])+layer.rnd_seed)
                        if layer.enabled:
                            if layer.name == 'Birds':
                                temp_rig = bpy.data.objects.get(layer.bird_rig_name)
                                if temp_rig != None and temp_rig.type == 'ARMATURE':
                                    for x in range(0,w):
                                        for y in range(0,h):
                                            if area_map[x][y] == '.':
                                                # Birds can appear here.
                                                bc = layer.bird_count
                                                for c in range(bc):
                                                    generateFlyingBird (context, c, x, y, layer, area_map, collection[grid[x][y]])
                                                layer.bird_count = bc
                                            else:
                                                # No birds should appear in this area.
                                                # We need to make an attempt to remove any previous birds by scanning for names of known objects.
                                                for c in range(layer.bird_count+10):
                                                    # Try to unlink prior count based candidates.
                                                    new_rig_name = returnItemName(layer.bird_rig_name, x, y, c)
                                                    try:
                                                        context.scene.objects.unlink(bpy.data.objects.get(new_rig_name))      # Unlink from scene if in scene.
                                                    except:
                                                        pass	# Already unlinked?
                                                    new_body_name = returnItemName(("%s_BODY" % layer.bird_rig_name),x, y, c)
                                                    try:
                                                        context.scene.objects.unlink(bpy.data.objects.get(new_body_name))      # Unlink from scene if in scene.
                                                    except:
                                                        pass	# Already unlinked?
                                                    path_bird_name = "path_bird_%s" % new_rig_name
                                                    try:
                                                        context.scene.objects.unlink(bpy.data.objects.get(path_bird_name))      # Unlink from scene if in scene.
                                                    except:
                                                        pass	# Already unlinked?

                            elif layer.name == 'Vehicles':
                                if ob.TileProcLayers_List[8].enabled:
                                    # Can only drive if roads have been generated.
                                    temp_rig = bpy.data.objects.get(layer.vehicle_rig_name)
                                    if temp_rig != None and temp_rig.type == 'ARMATURE':
                                        # Can only drive if a car rig has been specified.
                                        maze_map = maps[8]              # HARD CODED reference to slot #8 roads, bad coding.
                                        #maze_map = self.invertMap(maps[8])             # HARD CODED reference to slot #8 roads, bad coding.
                                        #self.printGrid(maze_map,"Road Maze Map [%s]." % layer.name)
                                        vc = layer.vehicle_count
                                        for c in range(vc):
                                            layer.vehicle_count = 1
                                            solution_map, solution_coordinates = self.solveMaze(maze_map)
                                            #self.printGrid(solution_map,"Drive Solution Map [%s]." % layer.name)
                                            if len(solution_coordinates) > 0:
                                                generateDrivingVehicle (context.scene, c, layer, area_map, maze_map, solution_map, solution_coordinates, grids[8], collection)
                                            else:
                                                print("bad coordinate set from solveMaze, vehicles generation not possible..?")
                                        layer.vehicle_count = vc
                        else:
                            if layer.name == 'Birds':
                                for x in range(0,w):
                                        for y in range(0,h):
                                            # No birds should appear in this area.
                                            # We need to make an attempt to remove any previous birds by scanning for names of known objects.
                                            for c in range(layer.bird_count+10):
                                                # Try to unlink prior count based candidates.
                                                new_rig_name = returnItemName(layer.bird_rig_name, x, y, c)
                                                try:
                                                    context.scene.objects.unlink(bpy.data.objects.get(new_rig_name))      # Unlink from scene if in scene.
                                                except:
                                                    pass	# Already unlinked?
                                                new_body_name = returnItemName(("%s_BODY" % layer.bird_rig_name),x, y, c)
                                                try:
                                                    context.scene.objects.unlink(bpy.data.objects.get(new_body_name))      # Unlink from scene if in scene.
                                                except:
                                                    pass	# Already unlinked?
                                                path_bird_name = "path_bird_%s" % new_rig_name
                                                try:
                                                    context.scene.objects.unlink(bpy.data.objects.get(path_bird_name))      # Unlink from scene if in scene.
                                                except:
                                                    pass	# Already unlinked?

                # Patch holes.
                if fill_amount > 0.0:
                    random.seed(int(ob["rnd_seed"]))
                    self.removeOrphans (grid, collection)                   # Throw away any bad connected tiles.
                    self.populateGridHoles (grid, collection)               # Fill initial holes.
                    #self.removeOrphans (grid, collection)                  # Throw away any bad connected tiles again.
                    #self.populateGridHoles (grid, collection)              # Try to fill the holes again.
                    self.patchWithForgivingConnections(grid, collection)    # Just fill the holes as best we can, no solution = grass.
                    lst = self.return2xAssetList(collection)
                    if len(lst):
                        n = random.randint(0,(len(lst)-1))
                        self.add2xAssets(lst[n], grid, collection)          # The index for the 2x size asset.
      
                self.addBorder (grid, collection)                           # If any border tiles are present, they will 'frame' the result.
                self.gridToObjects (context.scene, ob, grid, collection)
            else:
                to_console("Unable to generate, the tile list is empty?")
        return {'FINISHED'}

# Create operator to solidify a layer.   
class OBJECT_OT_solidify_layer(bpy.types.Operator):
    bl_label = "Solidify"
    bl_idname = "tileproc.solidify_layer"
    bl_description = "Reset area values for the selected layer to make the entire area solid."

    def invoke(self, context, event):
        ob = context.object
        if ob != None:
            collection = ob.TileProcLayers_List
            l = len(collection)
            if l > 1:
                # Reset the area to make this layer a solid.
                entry = collection[ob.TileProcLayers_List_Index]
                entry.area = 1.0
                entry.area_center_x = 0.0
                entry.area_center_y = 0.0
            else:
                to_console ("No layer detected.")
        return {'FINISHED'}
        
# Create operator to rename this object with the tileproc_ preifix.   
class OBJECT_OT_rename_to_TileProc(bpy.types.Operator):
    bl_label = "Rename To TileProc"
    bl_idname = "op.rename_to_tileproc"
    bl_description = "Click this button to rename this object with the tileproc__ prefix. This will make it a TileProc object."
    
    def invoke(self, context, event):
        ob = context.object
        if ob != None:
            TileProc_name = OBJECT_PREFIX + ob.name
            if len(TileProc_name) > MAX_NAME_SIZE:
                # Name too long.
                TileProc_name = TileProc_name[:MAX_NAME_SIZE]
            ob_source = bpy.data.objects.get(TileProc_name)
            if ob_source != None:
                # Hmm...already and object named like this.
                to_console ("Already an object named like [" + TileProc_name + "] rename manualy.")
            else:
                ob.name = TileProc_name
                ob.show_texture_space = True
                ob["width"] = 16
                ob["height"] = 16
                ob["rnd_seed"] = 311
                #ob["road_density"] = 0.80
                #ob["house_density"] = 0.25
                ob["filler_amount"] = 0.0
        return {'FINISHED'}

# Create operator to write the entries to a CSV text file.   
class OBJECT_OT_write_entries(bpy.types.Operator):
    """Write the list out to the desktop as a CSV text file sorted alpha by name."""
    bl_label = "Write"
    bl_idname = "tileproc.write_entries"
   
    def invoke(self, context, event):
        ob = context.object
        if ob != None:
            collection = ob.TileProc_List
            l = len(collection)
            if l > 0:
                render_name = os.path.basename(bpy.data.filepath)
                tileproc_output = os.path.expanduser("~/Desktop/%s.csv" % render_name)

                # Fix naming errors.
                for entry in collection:
                    n = entry.name.replace(("%s-" % ENTRY_NAME),("%s_" % ENTRY_NAME))
                    entry.name = n
                
                # Write file sorted by name.
                file = open(tileproc_output, "w")
                for entry in sorted(collection, key=attrgetter('name')):
                    s = "%s~%s~%s~%s~%s~%s~%s~%s~%s~%s\n" % (entry.name, entry.ob_name, entry.id_top, entry.id_bottom, entry.id_left, entry.id_right, entry.is_2x_asset, entry.is_buildable, entry.is_filler, entry.enabled)
                    file.write(s)
                file.close()
            else:
                to_console ("No entries in list to write.")
        return {'FINISHED'}

# Create operator to read the entries to a CSV text file.   
class OBJECT_OT_read_entries(bpy.types.Operator):
    """Read the scene named CSV file and replace the list with it's contents."""
    bl_label = "Read"
    bl_idname = "tileproc.read_entries"

    def str2bool(self, v):
      return v.lower() in ("yes", "true", "t", "1")
  
    def invoke(self, context, event):
        ob = context.object
        if ob != None:
            collection = ob.TileProc_List
            render_name = os.path.basename(bpy.data.filepath)
            tileproc_input = os.path.expanduser("~/Desktop/%s.csv" % render_name)
            lst = []
            try:
                with open(tileproc_input, "r") as file:
                    for line in file:
                        lst.append(line)
                    file.close()
            except:
                pass
            if len(lst):
                # Go ahead and blank the collection, we have new data to rebuild with.
                while len(collection):
                    collection.remove(0)

                for line in lst:
                    s = line.replace("\n","")       # Remove line feed.
                    ary = s.split("~")              # Split at tildae.
                    if len(ary) == 10:
                        collection.add()
                        collection[-1].name = ary[0]
                        collection[-1].ob_name = ary[1]
                        collection[-1].id_top = ary[2]
                        collection[-1].id_bottom = ary[3]
                        collection[-1].id_left = ary[4]
                        collection[-1].id_right = ary[5]
                        collection[-1].is_2x_asset = self.str2bool(ary[6])
                        collection[-1].is_buildable = self.str2bool(ary[7])
                        collection[-1].is_filler = self.str2bool(ary[8])
                        collection[-1].enabled = self.str2bool(ary[9])
                    else:
                        to_console ("Invalid split length, not our file.")
                ob.TileProc_List_Index = 0
            else:
                to_console ("Error reading file, it appears to be blank.")
        return {'FINISHED'}

# Create operator to unlink all tileobjects from the scene.   
class OBJECT_OT_delete_generation(bpy.types.Operator):
    """ Delete the current generation."""
    bl_label = "Delete"
    bl_idname = "tileproc.delete_generation"
   
    def invoke(self, context, event):
        plant_types = ['Trees', 'Bushes', 'Flowers']
        vehicle_types = ['Vehicles']
        ob = context.object
        if ob != None:
            w = int(ob["width"])
            h = int(ob["height"])
            for x in range(0,w):
                for y in range(0,h):
                    n = "%s_%s" % (ENTRY_NAME, ob.name[-6:])
                    tile_name = returnTileName(n, x, y) # Name of tile to be referenced.
                    temp_ob = bpy.data.objects.get(tile_name)
                    if temp_ob != None:
                        try:
                            context.scene.objects.unlink(temp_ob)
                        except:
                            pass
                    # Plant types are kind of like particles and need extensive looping to detect in scene.
                    for layer in ob.TileProcLayers_List:
                        if layer.name in plant_types:
                            for c in range(layer.item_count):
                                n = "%s_%s" % (ITEM_NAME, layer.name)
                                item_name = returnItemName(n, x, y, c) # Name of item to be created.
                                temp_ob_item = bpy.data.objects.get(item_name)
                                if temp_ob_item != None:
                                    try:
                                        context.scene.objects.unlink(temp_ob_item)          # Unlink from scene if in scene.
                                    except:
                                        # Already unlinked?
                                        pass
                        elif layer.name in vehicle_types:
                            # Unlink rig, mesh and supporting objects for vehicle type layers.
                            pass
        return {'FINISHED'}

# Create operator to add materials based upon tile type.   
class OBJECT_OT_choose_alternate(bpy.types.Operator):
    """ Swap selected tile with a candidate from the list of filtered tiles."""
    bl_label = "Choose Alternate Tile To Replace Selected Tile In Scene"
    bl_idname = "tileproc.choose_alternate"

    def list_update(self, context):
        #print("list update")
        pass
    
    def list_populate(self, context):
        lst = []
        result = []
        ob_tileproc = context.object
        if ob_tileproc != None:
            collection = ob_tileproc.TileProc_List
            l = len(collection)
            if l > 0:
                if len(context.selected_objects):
                    ob_to_find = context.selected_objects[0]
                    if ob_to_find != None:
                        #print("[%s] list_populate (%i, %i, %i, %i)" % (ob_to_find.name, self.is_match_top, self.is_match_bottom, self.is_match_left, self.is_match_right))
                        if ob_to_find.type == 'MESH':
                            found_index = -1
                            # Loop through the collection looking for ourself.
                            for (index, entry) in enumerate(collection):
                                temp_ob = bpy.data.objects.get(entry.ob_name)
                                if temp_ob != None:
                                    if temp_ob.type == 'MESH':
                                        if ob_to_find.data.name == temp_ob.data.name:
                                            # Found it, now make a list of other tiles with similar properties.
                                            found_index = index
                            if found_index != -1:
                                found_entry = collection[found_index]
                                # Loop through the collection looking for tiles like ourself, filtered by the checkboxes.
                                for (index, entry) in enumerate(collection):
                                    if entry.enabled:
                                        c = 0
                                        match_count = 0
                                        if self.is_match_top:
                                            match_count +=1
                                            if found_entry.id_top == entry.id_top: 
                                                c+=1
                                        if self.is_match_bottom:
                                            match_count +=1
                                            if found_entry.id_bottom == entry.id_bottom: 
                                                c+=1                                            
                                        if self.is_match_left:
                                            match_count +=1
                                            if found_entry.id_left == entry.id_left: 
                                                c+=1
                                        if self.is_match_right:
                                            match_count +=1
                                            if found_entry.id_right == entry.id_right: 
                                                c+=1    
                                        if c == match_count:
                                            lst.append(entry.ob_name)
            lst.sort(key=str.lower)
            for item in lst:
                result.append((item,item,'OUTLINER_OB_MESH'))   #Caption, key, icon..?
        else:
            result.append(('Invalid Context', 'invalid','QUESTION'))
        return result
  
    lst_alternate = bpy.props.EnumProperty(items=list_populate, description="A list of alternate compatible tiles for the location of the currently selected tile in the scene.", update=list_update)
    is_match_top = bpy.props.BoolProperty(name="Match On Top", description="Use to filter what appears in the candidate list.", default=True, subtype='NONE', update=list_populate)
    is_match_bottom = bpy.props.BoolProperty(name="Match On Bottom", description="Use to filter what appears in the candidate list.", default=True, subtype='NONE', update=list_populate)
    is_match_left = bpy.props.BoolProperty(name="Match On Left", description="Use to filter what appears in the candidate list.", default=True, subtype='NONE', update=list_populate)
    is_match_right = bpy.props.BoolProperty(name="Match On Right", description="Use to filter what appears in the candidate list.", default=True, subtype='NONE', update=list_populate)

    def draw(self,context):
        layout = self.layout
        box = layout.box()
        layout.label("TOP is in the 3D view Y-Positive direction.", icon='TRIA_UP')
        layout.label("RIGHT is in the 3D view X-Positive direction.", icon = 'TRIA_RIGHT')
        layout.prop(self,"lst_alternate",text='Tile Candidate', icon='OUTLINER_OB_MESH')
        box = layout.box()
        layout.label("Enable to require match to selected.")
        row = layout.row()
        col_1 = row.split(1.0)
        col_2 = row.split(1.0)
        col_3 = row.split(1.0)
        col_4 = row.split(1.0)
        col_1.prop(self, 'is_match_top', text = "Top")
        col_2.prop(self, 'is_match_bottom', text = "Bottom")
        col_3.prop(self, 'is_match_left', text = "Left")
        col_4.prop(self, 'is_match_right', text = "Right")

    def invoke(self, context, event):
        #Popup a dialog the user can interact with.
        wm = context.window_manager
        return wm.invoke_props_dialog(self)
        
    def execute(self, context):
        if len(context.selected_objects):
            ob_to_replace = context.selected_objects[0]
            if ob_to_replace.type == 'MESH':
                ob_candidate = bpy.data.objects.get(self.lst_alternate)
                if ob_candidate:
                    if ob_candidate.type == 'MESH':
                        # Swap the mesh datablock to finalize the replacement.
                        print ("swapping [%s] with [%s]" % (ob_to_replace.data.name, ob_candidate.data.name))
                        ob_to_replace.data  = ob_candidate.data
        return {'FINISHED'}

# Create operator to add materials based upon tile type.   
class OBJECT_OT_colorize_materials(bpy.types.Operator):
    """ Colorize the current materials from a base color."""
    bl_label = "Colorize Materials"
    bl_idname = "tileproc.colorize_materials"

    color = bpy.props.FloatVectorProperty(name="Color: ",default=(0.13,0.36,1.0), description="Base color to derive materials from.", subtype='COLOR')
    rnd_types = [
                ("00","eRacoon Original","original"),
                ("01","Pure Random","pure_random"),
                ("02","Preserve Hue","preserve_hue"),
                ("03","Preserve Hue And Value","preserve_hue_value"),
                ]
    rnd_type = bpy.props.EnumProperty(name="Color Type", description="The type for color generation.", default="03", items=rnd_types)
   
    def draw(self,context):
        layout = self.layout
        layout.prop(self,"rnd_type")
        #print(self.rnd_type)
        if self.rnd_type == "02" or self.rnd_type == "03":
            layout.prop(self,"color")

    def invoke(self, context, event):
        #Popup a dialog the user can interact with.
        wm = context.window_manager
        return wm.invoke_props_dialog(self)
        
    #def invoke(self, context, event):
    def execute(self, context):
        ob_tileproc = context.object
        if ob_tileproc != None:
            ob_mat_source = bpy.data.objects.get("aa_tileproc_material_container")          # HARD CODED reference to an object.
            if ob_mat_source != None:
                for ms in ob_mat_source.material_slots:
                    temp_mat = ms.material
                    #temp_mat = bpy.data.materials.get("grass")
                    if temp_mat != None:
                        if temp_mat.use_nodes:
                            nt = temp_mat.node_tree
                            for node in nt.nodes:
                                if node.name == 'Material Output':
                                    # review the input connection.
                                    inp = node.inputs[0]
                                    if inp.is_linked:
                                        from_node = inp.links[0].from_node
                                        # Blindly fetch input #0 as the diffuse color chit. (A chance to error out here...)
                                        match_color = from_node.inputs[0].default_value
                                        if self.rnd_type == "00":
                                            #print(node.name)
                                            # HARD CODED material name detection here.
                                            if temp_mat.name == 'dirt': c = (0.242,0.13,0.068,1)
                                            if temp_mat.name == 'grass': c = (0.242,0.429,0.068,1)
                                            if temp_mat.name == 'water': c = (0.15,0.479,0.651,1)
                                            if temp_mat.name == 'sand': c = (0.61,0.578,0.352,1)
                                            if temp_mat.name == 'road': c = (0.136,0.136,0.144,1)
                                            if temp_mat.name == 'concrete': c = (0.571,0.552,0.533,1)
                                            if temp_mat.name == 'station_roof': c = (0.366,0.198,0.109,1)
                                            if temp_mat.name == 'station_part-1': c = (0.791,0.209,0.016,1)
                                            if temp_mat.name == 'station_part-2': c = (0.402,0.371,0.223,1)
                                            if temp_mat.name == 'house_roof': c = (0.136,0.136,0.144,1)
                                            if temp_mat.name == 'house_wall': c = (0.571,0.552,0.533,1)
                                            if temp_mat.name == 'house_door': c = (0.223,0.12,0.063,1)
                                            if temp_mat.name == 'house_window': c = (0.359,0.409,0.462,1)
                                            if temp_mat.name == 'tunnel_lights': c = (1,1,1,1)
                                            if temp_mat.name == 'asphalt': c = (0.136,0.136,0.144,1)
                                        if self.rnd_type == "01":c = returnRandomColor()
                                        if self.rnd_type == "02":c = returnRandomColor(self.color)
                                        if self.rnd_type == "03":c = returnRandomColor(self.color, match_color)
                                        from_node.inputs[0].default_value = c
                                        temp_mat.diffuse_color = (c[0],c[1],c[2])       # Cycles viewport color.
                    else:
                        print("no mat")
            else:
                print ("no mat source")
        else:
            print ("context not valid.")
        return {'FINISHED'}  
              
# Create operator to add materials based upon tile type.   
class OBJECT_OT_materials_to_collection(bpy.types.Operator):
    """ Convert eRacoon single material to multiple Cycles materials."""
    bl_label = "Materials To Collection"
    bl_idname = "tileproc.materials_to_collection"
   
    def invoke(self, context, event):
        ob_tileproc = context.object
        if ob_tileproc != None:
            collection = ob_tileproc.TileProc_List
            l = len(collection)
            if l > 0:
                ob_mat_source = bpy.data.objects.get("aa_tileproc_material_container")      # HARD CODED reference to an object.
                if ob_mat_source != None:
                    for (index, entry) in enumerate(collection):
                        temp_ob = bpy.data.objects.get(entry.ob_name)
                        if temp_ob != None:
                            temp_ob.data.materials.clear()
                            mat_list = ['dirt','grass','water','sand','road','concrete']
                            if entry.ob_name.find("house") != -1: mat_list = ['dirt','grass','water','sand','road','concrete','house_roof','house_window','house_door','house_walk','house_walls']
                            if entry.ob_name.find("station") != -1: mat_list = ['dirt','grass','water','sand','road','concrete','station_roof','station_part-1','station_part-2']
                            for mat_name in mat_list:
                                mat = bpy.data.materials.get(mat_name)
                                if mat != None:
                                    # Add this material to the tile source object.
                                    temp_ob.data.materials.append(mat)
                            
                            # Try to determine material id face assignments from the UV bounding box of each face.
                            me = temp_ob.data
                            for face in me.polygons:
                                # Set extreme values to make sure the UVs take over and poulate these variables.
                                x1 = 10000.0
                                y1 = 10000.0
                                x2 = -10000.0
                                y2 = -10000.0
                                # Make a bounding box out of this face's UV coordinates.
                                for vertex_index in range(face.loop_start,(face.loop_start+face.loop_total)):
                                    uv = me.uv_layers.active.data[vertex_index].uv
                                    if uv[0] < x1: x1 = uv[0]
                                    if uv[1] < y1: y1 = uv[1]
                                    if uv[0] > x2: x2 = uv[0]
                                    if uv[1] > y2: y2 = uv[1]
                                if y2 > 0.54:
                                    # Resides in upper half of image map.
                                    if x2 > 0.666:
                                        # Grass
                                        ndx = 1
                                    elif x2 > 0.333:
                                        # Sand
                                        ndx = 3
                                    else:
                                        # Water
                                        ndx = 2
                                else:
                                    # Resides in lower half of image map.
                                    if x2 > 0.666:
                                        # Concrete
                                        ndx = 5
                                    elif x2 > 0.333:
                                        # Road
                                        ndx = 4
                                    else:
                                        # Dirt
                                        ndx = 0
                                face.material_index = ndx
                                print("mat Index %i for UV bounding box[%f,%f,%f,%f]" % (face.material_index,x1,y1,x2,y2))
            else:
                to_console ("No entries in list to locate.")
        return {'FINISHED'}
        
# Create operator to locate entry for selected tile based object.   
class OBJECT_OT_locate_entry(bpy.types.Operator):
    """ Jump to the entry in the list that references the selected object."""
    bl_label = "Locate"
    bl_idname = "tileproc.locate_in_collection"
   
    def invoke(self, context, event):
        ob_tileproc = context.object
        if ob_tileproc != None:
            collection = ob_tileproc.TileProc_List
            l = len(collection)
            if l > 0:
                if len(context.selected_objects):
                    ob_to_find = context.selected_objects[0]
                    if ob_to_find.type == 'MESH':
                        for (index, entry) in enumerate(collection):
                            temp_ob = bpy.data.objects.get(entry.ob_name)
                            if temp_ob != None:
                                if temp_ob.type != 'EMPTY':
                                    if ob_to_find.data.name == temp_ob.data.name:
                                        # Found it, set our list index.
                                        ob_tileproc.TileProc_List_Index = index
                                        break
            else:
                to_console ("No entries in list to locate.")
        return {'FINISHED'}
        
# Create operator to sort the entries   
class OBJECT_OT_sort_entries(bpy.types.Operator):
    bl_label = "Sort"
    bl_idname = "tileproc.sort_collection"
 
    def get_sort_key(self, entry):
        return entry.name
   
    def invoke(self, context, event):
        ob = context.object
        if ob != None:
            collection = ob.TileProc_List
            l = len(collection)
            if l > 1:
                for entry in sorted(collection, key=attrgetter('name')):
                    to_console (entry.name)
            else:
                to_console ("No entries in list to sort.")
        return {'FINISHED'}
        
# Create operator to add or remove entries to/from the Collection   
class OBJECT_OT_add_remove_String_Items(bpy.types.Operator):
    bl_label = "Add or Remove"
    bl_idname = "op.add_remove_tileproc_entry"
    add = bpy.props.BoolProperty(default = True)
    
    def invoke(self, context, event):
        add = self.add
        ob = context.object
        if ob != None:
            collection = ob.TileProc_List
            if add:
                # This adds at the end of the collection list.
                collection.add()
                l = len(collection)
                collection[-1].name= "%s-%i" % (ENTRY_NAME ,l)
            else:
                l = len(collection)
                if l > 1:
                    # This removes one item in the collection list function of index value
                    index = ob.TileProc_List_Index
                    collection.remove(index)
                else:
                    to_console ("Can not remove last item.")
        return {'FINISHED'}
        
# Create operator to to transfer the current selection to the list.
class OBJECT_OT_Tileproc_Get_Selection(bpy.types.Operator):
    bl_label = "Replace list with current selection"
    bl_description ="Replace the current tile list with a new list of the objects in the current selection. Pinnning this panel helps with this task."
    bl_idname = "tileproc.get_selection"

    def execute(self, context):
        global isBusy
        
        if isBusy == False:
            isBusy = True
            ob_tileproc = context.object
            if ob_tileproc != None:
                localScene = context.scene 
                # Proceed with acquiring targets from the scene selection.       
                ob_list = context.selected_objects
                                
                # Sort the list alphabetically and remove ourself from the list.
                from operator import itemgetter, attrgetter
                ob_list = sorted(ob_list,key=attrgetter('name'))
                sorted(ob_list, key=attrgetter('name'), reverse=True)
    
                if len(ob_list) > 0:
                    collection = ob_tileproc.TileProc_List
                    l = len(collection)
                    if l > 0:
                        #Remove previous data.
                        for n in range(l):
                            collection.remove(0)
                    else:
                        to_console("No list to remove, this is odd..?")

                    for ob in ob_list:
                        # Don't ever add self to list.
                        skip = False
                        if ob.name == ob_tileproc.name: skip = True
                        if skip == False:
                            # Only add objects that are not ourself.
                            collection.add()
                            l = len(collection)
                            collection[-1].name= "%s_%s-%i" % (ENTRY_NAME, ob.name ,l)  #(str(l)+ ENTRY_NAME + ob.name)
                            collection[-1].ob_name = (ob.name)
                    ob_tileproc.TileProc_List_Index = len(collection)-1  
                else:
                    to_console ("No items selected, list population not possible.")
            else:
                to_console ("Problem with context.object.")
            isBusy = False
        else:
            to_console("Still busy, operator [ot.get_selection] was not executed.")
        return {'FINISHED'}
        
def register():
    bpy.utils.register_class(OBJECT_OT_rename_to_TileProc)
    bpy.utils.register_class(OBJECT_OT_add_remove_String_Items)
    bpy.utils.register_class(OBJECT_OT_Tileproc_Get_Selection)
    bpy.utils.register_class(OBJECT_OT_generate_from_tiles)
    bpy.utils.register_class(OBJECT_OT_sort_entries)
    bpy.utils.register_class(OBJECT_OT_write_entries)
    bpy.utils.register_class(OBJECT_OT_read_entries)
    bpy.utils.register_class(OBJECT_OT_locate_entry)
    bpy.utils.register_class(OBJECT_OT_delete_generation)
    bpy.utils.register_class(OBJECT_OT_materials_to_collection)
    bpy.utils.register_class(OBJECT_OT_colorize_materials)
    bpy.utils.register_class(OBJECT_OT_solidify_layer)
    bpy.utils.register_class(OBJECT_OT_choose_alternate)

def unregister():
    bpy.utils.unregister_class(OBJECT_OT_rename_to_TileProc)
    bpy.utils.unregister_class(OBJECT_OT_add_remove_String_Items)
    bpy.utils.unregister_class(OBJECT_OT_Tileproc_Get_Selection)
    bpy.utils.unregister_class(OBJECT_OT_generate_from_tiles)
    bpy.utils.unregister_class(OBJECT_OT_sort_entries)
    bpy.utils.unregister_class(OBJECT_OT_write_entries)
    bpy.utils.unregister_class(OBJECT_OT_read_entries)
    bpy.utils.unregister_class(OBJECT_OT_locate_entry)
    bpy.utils.unregister_class(OBJECT_OT_delete_generation)
    bpy.utils.unregister_class(OBJECT_OT_materials_to_collection)
    bpy.utils.unregister_class(OBJECT_OT_colorize_materials)
    bpy.utils.unregister_class(OBJECT_OT_solidify_layer)
    bpy.utils.unregister_class(OBJECT_OT_choose_alternate)


