# AddOn Tile Proc (c) 2014 Atom

import bpy, math

#from .events import reviewTileProc

from .util import to_console

############################################################################
# Parameter Definitions That Can Be Animated And Appear In Panels
############################################################################
def updateTileProcParameter(self,context):
    try:
        scene = context.scene
    except:
        scene = None
    if scene != None:
        pass

class cls_TileProcLayer(bpy.types.PropertyGroup):
    enabled = bpy.props.BoolProperty(name="enabled", description="If active this layer will contribute in final generation.", default=True, subtype='NONE', update=updateTileProcParameter)
    rnd_seed = bpy.props.IntProperty(name="Layer Seed", description="The random seed added to the master seed to use for this layer.", default = 311, min = -4096, max = 4096)
    branch_rate = bpy.props.IntProperty(name="Branch Rate", description="Low values makes straight passages, high values make curvy ones.", default = 0, min = -10, max = 10)
    threshold = bpy.props.FloatProperty(name="Threshold", description="The chance that a long passage will have a hill.", default=0.65, min=0.0, max=1.0, update=updateTileProcParameter)
    area = bpy.props.FloatProperty(name="Area", description="The area of influence for this layer.", default=0.75, min=0.0, max=1.0, update=updateTileProcParameter)
    area_center_x = bpy.props.FloatProperty(name="Center X", description="The X position of the center of the area of influence.", default = 0.0, min = -1.0, max = 1.0)
    area_center_y = bpy.props.FloatProperty(name="Center Y", description="The Y position of the center of the area of influence.", default = 0.0, min = -1.0, max = 1.0)
    area_invert = bpy.props.BoolProperty(name="Invert Area", description="When enabled the area map will be inverted.", default=False, subtype='NONE', update=updateTileProcParameter)
    item_on = bpy.props.BoolProperty(name="Item On", description="When enabled this item is considered 'On'. With lamps, this allows the light to be turned on or off.", default=False, subtype='NONE', update=updateTileProcParameter)
    item_count = bpy.props.IntProperty(name="Count", description="Maximum number of items per tile.", default = 2, min = 0, max = 420)
    item_coverage = bpy.props.FloatProperty(name="Coverage", description="The coverage of items across a single tile.", default = 0.801, min = 0.01, max = 1.0)
    item_rnd_size = bpy.props.FloatProperty(name="RND Size", description="The amount of randomness to apply to the size.", default = 0.0, min = -1.0, max = 3.0)

    vehicle_count = bpy.props.IntProperty(name="Vehicle Count", description="Number of vehicle rigs to generate. Final scene is not very responsive if larger than 18. Use '/' key to solo rig", default = 2, min = 1, max = 256)
    vehicle_speed = bpy.props.IntProperty(name="Vehicle Speed", description="Number of frames it takes for a vehicle to travel across a tile. Lower is faster", default = 24, min = 5, max = 90)
    vehicle_rig_name = bpy.props.StringProperty(name="Vehicle Rig Name", description="The name of the armature object to clone for vehicle rig.")
    vehicle_lights = bpy.props.BoolProperty(name="Lights", description="If active any Lamp type objects that are children of the vehicle rig will be cloned as well.", default=True, subtype='NONE', update=updateTileProcParameter)
    vehicle_path = bpy.props.BoolProperty(name="Single Path", description="If active vehicle will travel along a single path. Otherwise a path for each tile is generated to facilitate bumpy roads.", default=True, subtype='NONE', update=updateTileProcParameter)

    bird_count = bpy.props.IntProperty(name="Bird Count", description="Number of bird rigs to generate. Final scene is not very responsive if larger than 18. Use '/' key to solo rig", default = 2, min = 1, max = 256)
    bird_speed = bpy.props.IntProperty(name="Bird Speed", description="Number of frames it takes the bird rig animation to completely travel around a circle.", default = 240, min = 24, max = 3600)
    bird_rig_name = bpy.props.StringProperty(name="Bird Rig Name", description="The name of the armature object to clone for bird rig.")
    hide_construction = bpy.props.BoolProperty(name="Hide Construction", description="If active the generated paths and armatures will be removed from the viewport.", default=True, subtype='NONE', update=updateTileProcParameter)

    cloud_count = bpy.props.IntProperty(name="Count", description="Maximum number of clouds.", default = 2, min = 0, max = 420)
    cloud_coverage = bpy.props.FloatProperty(name="Coverage", description="The coverage of items across a single tile.", default = 1.0, min = 0.01, max = 1.0)
    cloud_height = bpy.props.FloatProperty(name="Height", description="The height where clouds begin.", default = 5.0, min = 1.0, max = 30.0)
    cloud_rnd_size = bpy.props.FloatProperty(name="RND Size", description="The amount of randomness to apply to the size.", default = 0.0, min = -1.0, max = 3.0)
    cloud_axis_types = [
            ("00","X","x"),
            ("01","Y","y"),
    ]
    cloud_axis_type = bpy.props.EnumProperty(name="cloud_axis_type", description="Alter the roation of the clouds so they align with the X or Y axis.", default="00", items=cloud_axis_types, update=updateTileProcParameter)

    road_resolution = bpy.props.IntProperty(name="Road Resolution", description="Number of sub-divides within a single road section. Set higher if you use noise.", default = 1, min = 1, max = 32)
    road_noise = bpy.props.FloatProperty(name="Road Noise", description="The amount of bumpy-ness in a road segment. Leave 0.0 for flat roads.", default = 0.0, min = -0.5, max = 0.5)
    road_threshold = bpy.props.FloatProperty(name="Road Threshold", description="The flat contribution amount within the noise.", default = 0.0, min = 0.0, max = 1.0)
    lane_types = [
            ("00","Center","center"),
            ("01","Right","right"),
            ("02","Left","left"),
            ("03","Alternate","alternate"),
    ]
    road_lane_type = bpy.props.EnumProperty(name="road_lane_type", description="Starting lane type. Alternating lane types are altered for every other vehicle generated. So ODD #s are right and EVEN #s are left.", default="03", items=lane_types, update=updateTileProcParameter)

   # Border based properties.
    # Order should match collection creation order in TileProcLayers_new_source in ui.py.
    map_types = [
            ("00","None","none"),
            ("01","Dirt","dirt"),
            ("02","Sand","sand"),
            ("03","Water","water"),
            ("04","Grass","grass"),
            ("05","Hills","hills"),
            ("06","Dunes","dunes"),
            ("07","Rivers","rivers"),
            ("08","Canals","canals"),
            ("09","Roads","roads"),
            ("10","Houses","houses"),
            ("11","Trees","trees"),
            ("12","Bushes","bushes"),
            ("13","Flowers","flowers"),
            ("13","Vehicles","vehicles"),
            ("14","Street Assets","street_assets"),
            ("15","Advertising Assets","advertising_assets"),
            ("16","Birds","birds"),
            ("17","Clouds","clouds"),
            ("18","UFOs","ufos"),
    ]
    id_map_mask = bpy.props.EnumProperty(name="id_map_mask", description="When None use this layer's area. Otherwise use referenced layer as a mask or an avoidance map for this layer.", default="00", items=map_types, update=updateTileProcParameter)
 
               
class cls_TileProc(bpy.types.PropertyGroup):
    # The properties for this class which is referenced as an 'entry' below.
    ob_name = bpy.props.StringProperty(name="Object", description="The name of the object that these properties are associated with.")
    
    # NOTE: The operator.py code is HARD CODED to some of these entry numbers.
    # Do not insert other types or change the current order of the existing types or the code will no longer produce visualy accurate roads and rivers.
    # It is OK to append new types to the end of the existing list if you are extending the system.
    id_types = [
                ("00","None","None"),
                ("01","grass_low","Grass Low"),
                ("02","grass_high","Grass High"),
                ("03","road_low","Road Low"),
                ("04","road_high","Road High"),
                ("05","water_low","Water Low"),
                ("06","water_high","Water High"),
                ("07","sand","Sand"),
                ("08","dirt","Dirt"),
                ("09","asphalt","Asphalt"),
                ("10","tunnel","Tunnel"),
                ("11","sand_water_low","Sand Water Low"),
                ("12","sand_water_high","Sand Water High"),
                ("13","grass_water_low","Grass Water Low"),
                ("14","grass_water_high","Grass Water High"),
                ("15","canal_high","Canal High"),
                ("16","canal_low","Canal Low"),
                ("17","water_grass_low","Water Grass Low"),
                ("18","water_grass_high","Water Grass High"),
                ("19","water_sand_low","Water Sand Low"),
                ("20","water_sand_high","Water Sand High"),
                ("21","water_junction","Water Junction"),
                ("21","grass_asphalt","Grass Asphalt"),
                ("22","asphalt_grass","Asphalt Grass"),
                ("23","canal_high_water_low","Canal High Water Low"),
                ("24","house_connect","House Connect"),
                ("25","road_mid","Road Mid Connect"),
                ("26","tree","Tree"),
                ("27","bush","Bush"),
                ("28","flower","Flower"),
                ("29","vehicle_body","Vehicle Body"),
                ("30","vehicle_wheels","Vehicle Wheels"),
                ("31","street_asset","Street Assset"),
                ("32","advertising_asset","Advertising Asset"),
                ]
    id_top = bpy.props.EnumProperty(name="id_top", description="The ID type for the TOP part of the tile.", default="00", items=id_types, update=updateTileProcParameter)
    id_bottom = bpy.props.EnumProperty(name="id_bottom", description="The ID type for the BOTTOM part of the tile.", default="00", items=id_types, update=updateTileProcParameter)
    id_left = bpy.props.EnumProperty(name="id_left", description="The ID type for the LEFT part of the tile.", default="00", items=id_types, update=updateTileProcParameter)
    id_right = bpy.props.EnumProperty(name="id_right", description="The ID type for the RIGHT part of the tile.", default="00", items=id_types, update=updateTileProcParameter)
    is_2x_asset = bpy.props.BoolProperty(name="is_2x_asset", description="Activate if object takes up two grid spaces instead of one.", default=False, subtype='NONE', update=updateTileProcParameter)
    is_buildable = bpy.props.BoolProperty(name="is_buildable", description="Activate if object can be used to build 2x asset.", default=False, subtype='NONE', update=updateTileProcParameter)
    is_filler = bpy.props.BoolProperty(name="is_filler", description="Activate if object can be used to patch holes in grid.", default=False, subtype='NONE', update=updateTileProcParameter)
    is_plant = bpy.props.BoolProperty(name="is_plant", description="Activate if object can be used as plant.", default=False, subtype='NONE', update=updateTileProcParameter)
    is_vehicle_body = bpy.props.BoolProperty(name="is_vehicle_body", description="Activate if object can be used as a vehicle_body.", default=False, subtype='NONE', update=updateTileProcParameter)
    is_vehicle_wheels = bpy.props.BoolProperty(name="is_vehicle_wheels", description="Activate if object can be used as vehicle_wheels.", default=False, subtype='NONE', update=updateTileProcParameter)
    is_street_asset = bpy.props.BoolProperty(name="is_street_asset", description="Activate if object can be used as a street asset.", default=False, subtype='NONE', update=updateTileProcParameter)
    is_advertising_asset = bpy.props.BoolProperty(name="is_advertising_asset", description="Activate if object can be used as an advertising asset.", default=False, subtype='NONE', update=updateTileProcParameter)
    enabled = bpy.props.BoolProperty(name="enabled", description="Activate if object can be used in grid generation.", default=True, subtype='NONE', update=updateTileProcParameter)

    # Border based properties.
    border_types = [
            ("00","None","none"),
            ("01","Upper Left","ul"),
            ("02","Upper Right","ur"),
            ("03","Lower Left","ll"),
            ("04","Lower Right","lr"),
            ("05","Horizontal","hrz"),
            ("06","Vertical","vert"),
    ]
    id_border = bpy.props.EnumProperty(name="id_border", description="The border ID type for this tile. Just 1 tile or a tile set can be used to 'frame' the result. Only have 6 tiles defined for this to work. (UL,UR,LL,LR,H,V)", default="00", items=border_types, update=updateTileProcParameter)

# collection of property group classes that need to be registered on module startup
classes = [cls_TileProc, cls_TileProcLayer]

def register():
    print("tile_proc register")
    for cls in classes:
        bpy.utils.register_class(cls)

    # Add these properties to every object in the entire Blender system (muha-haa!!)
    bpy.types.Object.TileProc_List_Index = bpy.props.IntProperty(min = 0,default = 0,description="Internal value. Do not animate.")
    bpy.types.Object.TileProc_List = bpy.props.CollectionProperty(type=cls_TileProc,description="Internal list class.")
    # A second class just for layers.
    bpy.types.Object.TileProcLayers_List_Index = bpy.props.IntProperty(min = 0,default = 0,description="Internal value. Do not animate.")
    bpy.types.Object.TileProcLayers_List = bpy.props.CollectionProperty(type=cls_TileProcLayer,description="Internal list class.")

    # Modifer based properties.
    modifier_types = [
            ("00","None","none"),
            ("01","Preview","preview"),
            ("02","Render","render"),
    ]
    bpy.types.Object.TileProcMergeTiles = bpy.props.BoolProperty(name="Merge Tiles", description="When active final output will be a single mesh.", default=False, subtype='NONE', update=updateTileProcParameter)
    bpy.types.Object.TileProcApplyModifiers = bpy.props.EnumProperty(name="Apply Modifiers", description="When fetching datablock use this modifier type.", default="00", items=modifier_types, update=updateTileProcParameter)

def unregister():
    bpy.utils.unregister_module(__name__)