# AddOn Tile Proc (c) 2014 Atom

import bpy
from bpy.types import Menu, Panel, UIList
import threading, time
from bpy_extras.object_utils import AddObjectHelper

from .util import OBJECT_PREFIX
from .util import TILEPROC_OB_PREFIX
from .util import ENTRY_NAME

from .util import to_console
from .util import returnNameDroppedPrefix

from .auto_drive import returnWheelRotationFromDistance

#from .events import reviewTileProc

############################################################################
# Thread processing for parameters that are invalid to set in a DRAW context.
# By performing those operations in these threads we can get around the invalid CONTEXT error within a draw event.
# This is fairly abusive to the system and may result in instability of operation.
# Then again as long as you trap for stale data it just might work..?
# For best result keep the sleep time as quick as possible...no long delays here.
############################################################################
def TileProc_new_source(lock, passedSourceName, passedSleepTime):
    time.sleep(passedSleepTime) # Feel free to alter time in seconds as needed.   
    to_console("TileProc threading: TileProc_new_source")
    ob_source = bpy.data.objects.get(passedSourceName)
    if ob_source !=None:
            ob_source.show_name = True
            ob_source.show_texture_space = True
            # Populate the new entry in the collection list.
            collection = ob_source.TileProc_List
            collection.add()
            l = len(collection)
            collection[-1].name= "%s-%i" % (ENTRY_NAME ,l)
            to_console("TileProc threading: New entry established on [%s]." % passedSourceName)
    else:
        to_console("TileProc threading: Source not found [%s]." % passedSourceName) 
        
def TileProcLayers_new_source(lock, passedSourceName, passedSleepTime):
    time.sleep(passedSleepTime) # Feel free to alter time in seconds as needed.   
    to_console("TileProcLayers threading: TileProcLayers_new_source")
    ob_source = bpy.data.objects.get(passedSourceName)
    if ob_source !=None:
        # This is a new collection so create all the layers we need now.
        # Order of creation should match enum in cls_TileProcLayer of properties.py
        collection = ob_source.TileProcLayers_List
        collection.add()
        collection[-1].name= "Dirt"
        collection[-1].area = 0.58
        collection[-1].area_center_x = -8
        collection[-1].area_center_y = 8
        collection.add()
        collection[-1].name= "Sand"
        collection[-1].area = 0.68
        collection[-1].area_center_x = 8
        collection[-1].area_center_y = 8
        collection.add()
        collection[-1].name= "Water"
        collection[-1].area = 0.66
        collection[-1].area_center_x = 8
        collection[-1].area_center_y = -8
        collection.add()
        collection[-1].name= "Grass"
        collection[-1].area = 0.71
        collection[-1].area_center_x = -8
        collection[-1].area_center_y = -8
        collection.add()
        collection[-1].name= "Hills"
        collection[-1].area = 0.84
        collection.add()
        collection[-1].name= "Dunes"
        collection[-1].enabled = False
        collection.add()
        collection[-1].name= "Rivers"
        collection[-1].branch_rate = 3
        collection[-1].area = 1.0
        collection[-1].rnd_seed = 423
        collection.add()
        collection[-1].name= "Canals"
        collection[-1].branch_rate = -5
        collection[-1].area = 1.0
        collection[-1].rnd_seed = 288
        collection.add()
        collection[-1].name= "Roads"
        collection[-1].branch_rate = -3
        collection[-1].area = 1.0
        collection.add()
        collection[-1].name= "Houses"
        collection.add()
        collection[-1].name= "Trees"
        collection[-1].enabled = False
        collection[-1].rnd_seed = 311
        collection.add()
        collection[-1].name= "Bushes"
        collection[-1].enabled = False
        collection[-1].rnd_seed = 310
        collection.add()
        collection[-1].name= "Flowers"
        collection[-1].enabled = False
        collection[-1].rnd_seed = 309
        collection.add()
        collection[-1].name= "Vehicles"
        collection[-1].enabled = False
        collection[-1].vehicle_rig_name = "rig_car"
        collection.add()
        collection[-1].name= "Street Assets"
        collection[-1].enabled = False
        collection[-1].rnd_seed = 812
        collection.add()
        collection[-1].name= "Advertising Assets"
        collection[-1].enabled = False
        collection[-1].rnd_seed = 123
        collection[-1].area = 0.19
        collection.add()
        collection[-1].name= "Birds"
        collection[-1].enabled = False
        collection[-1].rnd_seed = 321
        collection.add()
        collection[-1].name= "Clouds"
        collection[-1].enabled = False
        collection[-1].rnd_seed = 812
        collection.add()
        collection[-1].name= "UFOs"
        collection[-1].enabled = False
        collection[-1].rnd_seed = 2112        
        ob_source.TileProcLayers_List_Index = 0

        to_console("TileProcLayers threading: New entry established on [%s]." % passedSourceName)
    else:
        to_console("TileProcLayers threading: Source not found [%s]." % passedSourceName) 

############################################################################
# PANEL code.
############################################################################   
supported_types = ['EMPTY'] 

# How to display list items.
class templateListLayers(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        area_types = ['Dirt','Sand','Grass', 'Water','Hills','Dunes']
        maze_types = ['Rivers','Canals','Roads']
        building_types = ['Houses']
        plant_types = ['Trees', 'Bushes', 'Flowers','Street Assets','Advertising Assets']
        vehicle_types = ['Vehicles', 'Birds']
        atmosphere_types = ['Clouds', 'UFOs']
        # draw_item must handle the three layout types... Usually 'DEFAULT' and 'COMPACT' can share the same code.
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            item_label = item.name
            if item.enabled:
                icon_enabled='CHECKBOX_HLT'
            else:
                 icon_enabled='CHECKBOX_DEHLT'

            if item.name in area_types:
                icon_name='FULLSCREEN'
            elif item.name in maze_types:
                icon_name='SNAP_PEEL_OBJECT'
            elif item.name in building_types:
                icon_name='UV_ISLANDSEL'
            elif item.name in plant_types:
                icon_name='STRANDS'
                layer = data.TileProcLayers_List[8]	# 8 is the layer slot where roads reside.
                if item.name == 'Street Assets':
                    # Are roads enabled?
                    if layer.enabled == False:
                        item_label = "%s (Roads must be Enabled for street assets)" % item.name
            elif item.name in vehicle_types:
                icon_name='AUTO'
                if item.name == 'Vehicles':
                    layer = data.TileProcLayers_List[8]	# 8 is the layer slot where roads reside.
                    # Are roads enabled?
                    if layer.enabled == False:
                        item_label = "%s (Roads must be Enabled for vehicles)" % item.name
                if item.name == 'Birds':
                    icon_name = 'BOIDS'
            elif item.name in atmosphere_types:
                if item.name == 'Clouds':
                    icon_name = 'BRUSH_BLOB'
                elif item.name == 'UFOs':	
                    icon_name = 'MESH_CONE'
            else:
                icon_name='QUESTION'

            split = layout.split()
            col = split.column()
            row = col.row(align=True)
            row.prop(item,"enabled",text='',icon=icon_enabled, emboss=False)
            row.label(text=item_label, icon=icon_name)
        # 'GRID' layout type should be as compact as possible (typically a single icon!).
        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            layout.label(text="", icon_value=icon)
            
class templateListItems(UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):
        # draw_item must handle the three layout types... Usually 'DEFAULT' and 'COMPACT' can share the same code.
        if self.layout_type in {'DEFAULT', 'COMPACT'}:
            layout.label(text=item.name, text_ctxt="", translate=False, icon_value=4)
        # 'GRID' layout type should be as compact as possible (typically a single icon!).
        elif self.layout_type in {'GRID'}:
            layout.alignment = 'CENTER'
            layout.label(text="", icon_value=icon)

class OBJECT_PT_TileProcLayers(bpy.types.Panel):
    bl_label = "Tile Proc: Layers"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"

    @classmethod
    def poll(cls, context):
        result = False
        if context != None:
            ob = context.object
            if ob != None:
                if ob.type in supported_types:
                    result = True
        return result

    def draw(self, context):
        try:
            # If the name space does not exist we get an error.
            ns = bpy.app.driver_namespace["returnWheelRotationFromDistance"]
        except:
            # Try to establish the name space which manages our wheel rotations.
            bpy.app.driver_namespace["returnWheelRotationFromDistance"] = returnWheelRotationFromDistance
        
        area_types = ['Dirt','Sand','Water', 'Grass,','Hills','Dunes']
        maze_types = ['Rivers','Canals','Roads']
        plant_types = ['Trees', 'Bushes', 'Flowers','Street Assets','Advertising Assets']
        vehicle_types = ['Vehicles', 'Birds']
        atmosphere_types = ['Clouds', 'UFOs']
        if context != None:
            ob = context.object
            if ob != None:
                if ob.type in supported_types:
                    layout = self.layout
                    try:
                        # Looks like all is good, we can proceed.
                        ob = context.object
                        can_proceed = True
                    except:
                        # We got an error, we are in a crash/rendering state.
                        # Skip this update.
                        can_proceed = False
                    if can_proceed == True:
                        if ob != None:
                            if ob.type in supported_types:
                                l = len(OBJECT_PREFIX)
                                if ob.name[:l] == OBJECT_PREFIX:
                                    try:
                                        l = len(ob.TileProc_List)
                                    except:
                                        l = 0
                                    if l > 0:
                                        # Related items go in this box.
                                        box = layout.box()
                                        #box = box.box()
                                        try:
                                            l2 = len(ob.TileProcLayers_List)
                                        except:
                                            l2 = 0
                                        if l2 == 0:
                                            # We have no collections so we have to add one.
                                            # Launch a thread to set the remaining values that would generate an error if issued now.
                                            lock = threading.Lock()
                                            lock_holder = threading.Thread(target=TileProcLayers_new_source, args=(lock,ob.name,0.02), name='TileProcLayers_New_Source')
                                            lock_holder.setDaemon(True)
                                            lock_holder.start()
                                            return
                                        else:
                                            row = box.row()
                                            row.template_list("templateListLayers","layers",ob, "TileProcLayers_List", ob, "TileProcLayers_List_Index",2,2)
                                            col = row.column(align=True)
                                            col.operator("tileproc.solidify_layer", icon="MESH_GRID", text="")
                                            #col.operator("tileproc.materials_to_collection", icon="HAND", text="")	 # Only needed to convert old style eRacoon materials to a Cycles material per unique face color.
                                            col.operator("tileproc.colorize_materials", icon="COLOR", text="")
                                            col.operator("tileproc.delete_generation", icon="CANCEL", text="")
                                            

                                            #col.operator("op.add_remove_tileproc_entry", icon="ZOOMOUT", text="").add = False	# This show a minus sign button
                                            layer_entry = ob.TileProcLayers_List[ob.TileProcLayers_List_Index]
                                            box2 = box.box() 
                                            
                                            split = box2.split()
                                            col = split.column()
                                            row = col.row(align=True)
                                            #row.label(" Tile:", icon='SPLITSCREEN')
                                            #row.prop(entry, "name", text = '')
                                            #row.prop(layer_entry, 'enabled', text = "Enabled")
                                            row.label("%s Settings:" % layer_entry.name, icon='SETTINGS')
                                            row.prop(layer_entry, 'rnd_seed')
                                            
                                            if layer_entry.name in maze_types:
                                                split = box2.split()
                                                col = split.column()
                                                col.label("Maze Controls:", icon='SNAP_PEEL_OBJECT')
                                                row = col.row(align=True)
                                                row.prop(layer_entry, 'branch_rate')
                                                row.prop(layer_entry, 'threshold')
                                            elif layer_entry.name in plant_types:
                                                split = box2.split()
                                                col = split.column()
                                                #col.label("Item Controls:", icon='SNAP_PEEL_OBJECT')
                                                row = col.row(align=True)
                                                row.prop(layer_entry, 'item_count')
                                                row.prop(layer_entry, 'item_coverage')
                                                row.prop(layer_entry, 'item_rnd_size')
                                                #row.prop(layer_entry, 'item_on', text = "")
                                            elif layer_entry.name in vehicle_types:
                                                split = box2.split()
                                                col = split.column()
                                                if layer_entry.name == 'Vehicles':
                                                    # Let's let the icon offer a little feedback.
                                                    if bpy.data.objects.get(layer_entry.vehicle_rig_name) == None:
                                                        col.prop(layer_entry, 'vehicle_rig_name', icon='QUESTION')
                                                    else:
                                                        col.prop(layer_entry, 'vehicle_rig_name', icon='OBJECT_DATAMODE')
                                                    split = box2.split()
                                                    col = split.column()
                                                    row = col.row(align=True)
                                                    row.prop(layer_entry, 'vehicle_count')
                                                    row.prop(layer_entry, 'vehicle_speed')
                                                    row.prop(layer_entry, 'road_lane_type', text='Lane')
                                                    row = col.row(align=True)
                                                    row.prop(layer_entry, 'vehicle_lights')
                                                    row.prop(layer_entry, 'vehicle_path')
                                                    if layer_entry.vehicle_path == False:
                                                        row = col.row(align=True)
                                                        row.prop(layer_entry, 'road_resolution')
                                                        row.prop(layer_entry, 'road_noise')
                                                        row.prop(layer_entry, 'road_threshold')
                                                elif layer_entry.name == 'Birds':
                                                    # Let's let the icon offer a little feedback.
                                                    if bpy.data.objects.get(layer_entry.bird_rig_name) == None:
                                                        col.prop(layer_entry, 'bird_rig_name', icon='QUESTION')
                                                    else:
                                                        col.prop(layer_entry, 'bird_rig_name', icon='OBJECT_DATAMODE')
                                                    split = box2.split()
                                                    col = split.column()
                                                    row = col.row(align=True)
                                                    row.prop(layer_entry, 'bird_count')
                                                    row.prop(layer_entry, 'bird_speed')
                                                    row.prop(layer_entry, 'hide_construction')
                                            elif layer_entry.name in atmosphere_types:
                                                split = box2.split()
                                                col = split.column()
                                                row = col.row(align=True)
                                                row.prop(layer_entry, 'cloud_count')
                                                row.prop(layer_entry, 'cloud_height')	
                                                row.prop(layer_entry, 'cloud_rnd_size')	
                                                row.prop(layer_entry, 'cloud_axis_type',text="")	
                                            box1 = box2.box()
                                            split = box2.split()
                                            col = split.column()
                                            row = col.row(align=True)
                                            row.prop(layer_entry, 'area_invert')
                                            row.prop(layer_entry, 'id_map_mask', text='Mask Layer')
                                                
                                            #box2.prop(layer_entry, 'id_map_mask', text='Mask Layer')
                                            if layer_entry.id_map_mask == '00':
                                                # Display area controls.
                                                box2.prop(layer_entry, 'area')
                                                split = box2.split()
                                                col = split.column()
                                                row = col.row(align=True)
                                                row.prop(layer_entry, 'area_center_x')
                                                row.prop(layer_entry, 'area_center_y')

class OBJECT_PT_TileProcTiles(bpy.types.Panel):
    bl_label = "Tile Proc: Tiles"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"

    @classmethod
    def poll(cls, context):
        result = False
        if context != None:
            ob = context.object
            if ob != None:
                if ob.type in supported_types:
                    result = True
        return result

    def draw(self, context):
        #area_types = ['Dirt','Sand','Water', 'Grass,','Hills','Dunes']
        #maze_types = ['Rivers','Canals','Roads']
        if context != None:
            ob = context.object
            if ob != None:
                if ob.type in supported_types:
                    layout = self.layout
                    try:
                        # Looks like all is good, we can proceed.
                        ob = context.object
                        can_proceed = True
                    except:
                        # We got an error, we are in a crash/rendering state.
                        # Skip this update.
                        can_proceed = False
                    if can_proceed == True:
                        if ob != None:
                            if ob.type in supported_types:
                                l = len(OBJECT_PREFIX)
                                if ob.name[:l] == OBJECT_PREFIX:
                                    try:
                                        l = len(ob.TileProc_List)
                                    except:
                                        l = 0
                                    if l > 0:
                                        # Related items go in this box.
                                        box = layout.box()
                                        # Display list of targets in panel.

                                        # Look for None entries in the mapping fields to indicate trouble.
                                        icon_name = 'COLOR_GREEN'
                                        problem_entry = ''
                                        for entry in ob.TileProc_List:
                                            #entry = ob.TileProc_List[n]
                                            if entry.id_top == "00":
                                                icon_name = 'COLOR_RED'
                                                problem_entry = entry.name
                                                break
                                            if entry.id_bottom == "00":
                                                icon_name = 'COLOR_RED'
                                                problem_entry = entry.name
                                                break
                                            if entry.id_left == "00":
                                                icon_name = 'COLOR_RED'
                                                problem_entry = entry.name
                                                break
                                            if entry.id_right == "00":
                                                icon_name = 'COLOR_RED'
                                                problem_entry = entry.name
                                                break
                                            if bpy.data.objects.get(entry.ob_name) == None:
                                                icon_name = 'COLOR_BLUE'
                                                problem_entry = entry.name
                                                break
                                        try:
                                            cur_ndx = ob.TileProc_List_Index
                                        except:
                                            cur_ndx = -1
                                        box.label("Managed Tiles: %i of %i %s" % (cur_ndx, l, problem_entry),icon = icon_name)

                                        row = box.row()
                                        row.template_list("templateListItems","tiles",ob, "TileProc_List", ob, "TileProc_List_Index",6,6)
                                        
                                        # Display Add/Remove Operato.r
                                        col = row.column(align=True)
                                        col.operator("op.add_remove_tileproc_entry", icon="ZOOMIN", text="").add = True    # This show a plus sign button
                                        col.operator("op.add_remove_tileproc_entry", icon="ZOOMOUT", text="").add = False	# This show a minus sign button
                                        col.separator()
                                        col.separator()
                                        col.operator("tileproc.get_selection", icon="BORDER_RECT", text="")		#RESTRICT_SELECT_OFF
                                        #col.operator("tileproc.sort_entries", icon="SORTALPHA", text="")
                                        col.operator("tileproc.write_entries", icon="LIBRARY_DATA_DIRECT", text="")
                                        col.operator("tileproc.read_entries", icon="TEXT", text="")
                                        #col.operator("tileproc.materials_to_collection", icon="HAND", text="")	 # Only needed to convert old style eRacoon materials to a Cycles material per unique face color.
                                        #col.operator("tileproc.delete_generation", icon="CANCEL", text="")
                                        col.operator("tileproc.choose_alternate", icon="FILE_REFRESH", text="")
                                        col.operator("tileproc.locate_in_collection", icon="ZOOM_SELECTED", text="")
 
                                        if  ob.TileProc_List_Index >= len(ob.TileProc_List):
                                            # If you use an arrow down when the list box has focus you can end up here.
                                            #ob.Selection_List_Index = len(ob.Selection_List)-1
                                            pass
                                        else:
                                            # Use texture space to highlight the selected tile object in the scene.
                                            # First turn off texture space for all tiles in the list.
                                            for entry in ob.TileProc_List:
                                                temp_ob = bpy.data.objects.get(entry.ob_name)
                                                if temp_ob != None:
                                                    temp_ob.show_texture_space = False
                                                    
                                            entry = ob.TileProc_List[ob.TileProc_List_Index]
                                            # Turn on the texure space for the currently selected tile in the list.
                                            temp_ob = bpy.data.objects.get(entry.ob_name)
                                            if temp_ob != None:
                                                temp_ob.show_texture_space = True

                                            box = box.box() 
                                            split = box.split()
                                            col = split.column()
                                            row = col.row(align=True)
                                            row.label(" Tile:", icon='SPLITSCREEN')
                                            row.prop(entry, "name", text = '')
                                            row = box.row()
                                            # Let's let the icon offer a little feedback.
                                            if bpy.data.objects.get(entry.ob_name) == None:
                                                box.prop(entry, "ob_name", icon='QUESTION')
                                            else:
                                                box.prop(entry, "ob_name", icon='OBJECT_DATAMODE')
            
                                            layout.separator()
                                            box2 = box.box() 
                                            box2.prop(entry, 'enabled', text = "Enabled")
                                            box2.prop(entry, 'id_top', text = "Top")
                                            box2.prop(entry, 'id_bottom', text = "Bottom")
                                            box2.prop(entry, 'id_left', text = "Left")
                                            box2.prop(entry, 'id_right', text = "Right")
                                            box2.box()
                                            box2.prop(entry, 'id_border', text = "Border")
                                            
                                            row = box2.row()
                                            col_1 = row.split(1.0)
                                            col_2 = row.split(1.0)
                                            col_3 = row.split(1.0)
                                            col_4 = row.split(1.0)
                                            col_5 = row.split(1.0)
                                            col_6 = row.split(1.0)
                                            col_1.prop(entry, 'is_2x_asset', text = "2x Asset")
                                            col_2.prop(entry, 'is_buildable', text = "Buildable")
                                            col_3.prop(entry, 'is_filler', text = "Filler")
                                            col_4.prop(entry, 'is_plant', text = "Plant")
                                            col_5.prop(entry, 'is_street_asset', text = "Street")
                                            col_6.prop(entry, 'is_advertising_asset', text = "Ad")
       
class OBJECT_PT_TileProcGenerate(bpy.types.Panel):
    bl_label = "Tile Proc: Generate"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "object"

    @classmethod
    def poll(cls, context):
        result = False
        if context != None:
            ob = context.object
            if ob != None:
                if ob.type in supported_types:
                    result = True
        return result

    def draw(self, context):
        #area_types = ['Dirt','Sand','Water', 'Grass,','Hills','Dunes']
        #maze_types = ['Rivers','Canals','Roads']
        if context != None:
            ob = context.object
            if ob != None:
                if ob.type in supported_types:
                    layout = self.layout
                    try:
                        # Looks like all is good, we can proceed.
                        ob = context.object
                        can_proceed = True
                    except:
                        # We got an error, we are in a crash/rendering state.
                        # Skip this update.
                        can_proceed = False
                    if can_proceed == True:
                        if ob != None:
                            if ob.type in supported_types:
                                l = len(OBJECT_PREFIX)
                                if ob.name[:l] == OBJECT_PREFIX:
                                    try:
                                        l = len(ob.TileProc_List)
                                    except:
                                        l = 0
                                    if l > 0:
                                        # Related items go in this box.
                                        box = layout.box()
                                        split = box.split()
                                        col = split.column()
                                        col.label("Grid Size:")
                                        row = col.row(align=True)
                                        row.prop(ob, '["width"]', text = 'Width')
                                        row.prop(ob, '["height"]', text='Height')
                                        split = box.split()
                                        col = split.column()
                                        row = col.row(align=True)
                                        row.operator("op.generate_from_tiles", icon="MOD_BUILD", text="Generate")
                                        row.prop(ob, '["filler_amount"]', text='Fill Gaps')
                                        row.prop(ob, '["rnd_seed"]', text='Seed')
                                    else:
                                        # We have no collections so we have to add one.
                                        # But Blender has a WriteID lock in place at this time.
                                        # So we launch a thread that will run in a very, very short time from now.
                                        # Meanwhile we simply exit because the def the thread calls will do the work we intended to do here.
                                        #ob.TileProc_List.add()
                                        
                                        # Launch a thread to set the remaining values that would generate an error if issued now. (listed below)
                                        lock = threading.Lock()
                                        lock_holder = threading.Thread(target=TileProc_new_source, args=(lock,ob.name,0.02), name='TileProc_New_Source')
                                        lock_holder.setDaemon(True)
                                        lock_holder.start()
                                else:
                                    # Common to end up here for other non-TileProc objects.
                                    layout.label("Not a TileProc object yet.",icon='INFO')  
                                    layout.operator("op.rename_to_tileproc", icon="SORTALPHA", text="(rename with '%s' prefix to enable)" % OBJECT_PREFIX)
                            else:
                                # Unsupported object type.
                                layout.label("Selected object can not be a TileProc object.",icon='ERROR') 
                        else:
                            to_console("We fetch None for the context object without throwing an error..?")
                    else:
                        to_console("ERROR: Can not proceed..?")
        else:
            # This can happen sometimes after a render.
            to_console ("TileProc was given an invalid context, imagine that..")
            self.layout.label("TileProc was given an invalid context.",icon='HELP')

###########################################################################
# Menu code.
###########################################################################
def install_tileproc(self, context):
    # Time to actually create a TileProc object.
    ob = bpy.data.objects.new("%sManager" % OBJECT_PREFIX,None)
    if ob != None:
        context.scene.objects.link(ob)
        result = "TileProc [%s] added to the scene." % ob.name
        ob.show_axis = True
        ob["width"] = 16
        ob["height"] = 16
        ob["rnd_seed"] = 311
        ob["road_density"] = 0.80
        ob["house_density"] = 0.25
        ob["filler_amount"] = 0.0
        
        # Add the first entry to the collection.
        collection = ob.TileProc_List
        collection.add()
        l = len(collection)
        collection[-1].name= ("%s-%i" % (ENTRY_NAME,l))
    else:
        result = "Failed to create a new TileProc object." 
    to_console(result)				

    
class OBJECT_OT_add_tileproc(bpy.types.Operator, AddObjectHelper):
    """TileProc"""
    bl_idname = "ob.add_tileproc"
    bl_label = "Tile Proc"
    bl_description = "Generate a grid based layout from a tile source pool."
    bl_options = {'REGISTER', 'UNDO'}
    
    def execute(self, context):
        install_tileproc(self, context)
        return {'FINISHED'}

def add_tileproc(self, context):
    self.layout.operator(
        OBJECT_OT_add_tileproc.bl_idname,
        text="TileProc",
        icon="SPLITSCREEN")	#FACESEL_HLT
           
def register():
    bpy.utils.register_class(OBJECT_PT_TileProcGenerate)
    bpy.utils.register_class(OBJECT_PT_TileProcLayers)
    bpy.utils.register_class(OBJECT_PT_TileProcTiles)
    bpy.utils.register_class(templateListItems)
    bpy.utils.register_class(templateListLayers)
    bpy.utils.register_class(OBJECT_OT_add_tileproc)
    bpy.types.INFO_MT_mesh_add.append(add_tileproc)

def unregister():
    bpy.utils.unregister_class(OBJECT_PT_TileProcGenerate)
    bpy.utils.unregister_class(OBJECT_PT_TileProcLayers)
    bpy.utils.unregister_class(OBJECT_PT_TileProcTiles)
    bpy.utils.unregister_class(templateListItems)
    bpy.utils.unregister_class(templateListLayers)
    bpy.utils.unregister_class(OBJECT_OT_add_tileproc)
    bpy.types.INFO_MT_mesh_add.remove(add_tileproc)

